#!/usr/local/bin/python
# encoding: utf-8
"""
*import the ATLAS stream into the marshall*

:Author:
    David Young
"""
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
from astrocalc.times import conversions
from fundamentals.mysql import writequery
from marshallEngine.feeders.atlas.lightcurve import generate_atlas_lightcurves
from datetime import datetime, date, time, timedelta


class data(basedata):
    """
    *Import the ATLAS transient data into the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``dbConn`` -- the marshall database connection
    - ``settings`` -- the settings dictionary

    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a data object, use the following:

    ```python
    from marshallEngine.feeders.atlas.data import data
    ingester = data(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).ingest(withinLastDays=withInLastDay)   
    ```

    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn

        self.fsTableName = "fs_atlas"
        self.survey = "ATLAS"

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments**

        - ``withinLastDays`` -- within the last number of days. *Default: 50*

        """
        self.log.debug('starting the ``ingest`` method')

        timelimit = datetime.now() - timedelta(days=int(withinLastDays))
        timelimit = timelimit.strftime("%Y-%m-%d")

        csvDicts = self.get_csv_data(
            url=self.settings["atlas urls"]["summary csv"] + f"?followup_flag_date__gte={timelimit}"
        )

        self._clean_data_pre_ingest(
            surveyName="ATLAS", withinLastDays=withinLastDays)

        self._import_to_feeder_survey_table()
        self.insert_into_transientBucket(updateTransientSummaries=False)

        sqlQuery = """call update_fs_atlas_forced_phot()""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        self.fsTableName = "fs_atlas_forced_phot"
        self.survey = "ATLAS FP"

        sqlQuery = """CALL update_transientBucket_atlas_sources()""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        self.insert_into_transientBucket(importUnmatched=False)

        # UPDATE THE ATLAS SPECIFIC FLUX SPACE LIGHTCURVES
        generate_atlas_lightcurves(
            log=self.log,
            dbConn=self.dbConn,
            settings=self.settings
        )

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.debug('completed the ``ingest`` method')
        return None

    def _clean_data_pre_ingest(
            self,
            surveyName,
            withinLastDays=False):
        """*clean up the list of dictionaries containing the ATLAS data, pre-ingest*

        **Key Arguments**

        - ``surveyName`` -- the ATLAS survey name
        -  ``withinLastDays`` -- the lower limit of observations to include (within the last N days from now). Default *False*, i.e. no limit


        **Return**

        - ``dictList`` -- the cleaned list of dictionaries ready for ingest


        **Usage**

        To clean the data from the ATLAS survey:

        ```python
        dictList = ingesters._clean_data_pre_ingest(surveyName="ATLAS")
        ```

        Note you will also be able to access the data via ``ingester.dictList``

        """
        self.log.debug('starting the ``_clean_data_pre_ingest`` method')

        self.dictList = []

        # CALC MJD LIMIT
        if withinLastDays:
            mjdLimit = now(
                log=self.log
            ).get_mjd() - float(withinLastDays)

        # CONVERTER TO CONVERT MJD TO DATE
        converter = conversions(
            log=self.log
        )

        for row in self.csvDicts:
            # IF NOW IN THE LAST N DAYS - SKIP
            flagMjd = converter.ut_datetime_to_mjd(
                utDatetime=row["followup_flag_date"])

            if withinLastDays and (float(row["earliest_mjd"]) < mjdLimit and float(flagMjd) < mjdLimit):
                continue

            # MASSAGE THE DATA IN THE INPUT FORMAT TO WHAT IS NEEDED IN THE
            # FEEDER SURVEY TABLE IN THE DATABASE
            target = row["target"]
            diff = row["diff"]
            ref = row["ref"]
            targetImageURL = None
            refImageURL = None
            diffImageURL = None

            if target:
                mjdStr = str(int(float(target.split("_")[1])))
                if target:
                    iid, mjdString, diffId, ippIdet, type = target.split('_')
                    targetImageURL = "https://star.pst.qub.ac.uk/sne/atlas4/site_media/images/data/atlas4/" % locals() + '/' + \
                        mjdStr + '/' + target + '.jpeg'
                    objectURL = "https://star.pst.qub.ac.uk/sne/atlas4/candidate/" + iid

            if ref:
                mjdStr = str(int(float(ref.split("_")[1])))
                if ref:
                    iid, mjdString, diffId, ippIdet, type = ref.split('_')
                    refImageURL = "https://star.pst.qub.ac.uk/sne/atlas4/site_media/images/data/atlas4/" % locals() + '/' + \
                        mjdStr + '/' + ref + '.jpeg'
                    objectURL = "https://star.pst.qub.ac.uk/sne/atlas4/candidate/" + iid

            if diff:
                mjdStr = str(int(float(diff.split("_")[1])))
                if diff:
                    iid, mjdString, diffId, ippIdet, type = diff.split('_')
                    diffImageURL = "https://star.pst.qub.ac.uk/sne/atlas4/site_media/images/data/atlas4/" % locals() + '/' + \
                        mjdStr + '/' + diff + '.jpeg'
                    objectURL = "https://star.pst.qub.ac.uk/sne/atlas4/candidate/" + iid

            discDate = converter.mjd_to_ut_datetime(
                mjd=row["earliest_mjd"], sqlDate=True)

            thisDictionary = {}
            thisDictionary["candidateID"] = row["name"]
            thisDictionary["ra_deg"] = row["ra"]
            thisDictionary["dec_deg"] = row["dec"]
            thisDictionary["mag"] = row["earliest_mag"]
            thisDictionary["observationMJD"] = row["earliest_mjd"]
            thisDictionary["filter"] = row["earliest_filter"]
            thisDictionary["discDate"] = discDate
            thisDictionary["discMag"] = row["earliest_mag"]
            thisDictionary["suggestedType"] = row["object_classification"]
            thisDictionary["targetImageURL"] = targetImageURL
            thisDictionary["refImageURL"] = refImageURL
            thisDictionary["diffImageURL"] = diffImageURL
            thisDictionary["objectURL"] = objectURL

            self.dictList.append(thisDictionary)

        self.log.debug('completed the ``_clean_data_pre_ingest`` method')
        return self.dictList

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*Generate the force-photometry lightcurves for ATLAS sources found in the marshall database*

:Author:
    David Young
"""
from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from past.utils import old_div
import sys
import os
# SUPPRESS MATPLOTLIB WARNINGS
import warnings
warnings.filterwarnings("ignore")
import matplotlib as mpl
import matplotlib.pyplot as plt
import math
import numpy as np
from matplotlib import dates
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from datetime import datetime
from operator import itemgetter
from fundamentals.stats import rolling_window_sigma_clip
import matplotlib.ticker as mtick
from astrocalc.times import conversions
from fundamentals import fmultiprocess
from fundamentals.mysql import database, readquery, writequery


def generate_atlas_lightcurves(
        dbConn,
        log,
        settings):
    """generate all atlas FP lightcurves (clipped and stacked)

    **Key Arguments**

    - ``dbConn`` -- mysql database connection
    - ``log`` -- logger
    - ``settings`` -- settings for the marshall.

    ```python
    from marshallEngine.feeders.atlas.lightcurve import generate_atlas_lightcurves
    generate_atlas_lightcurves(
        log=log,
        dbConn=dbConn,
        settings=settings
    )
    ```
    """
    log.debug('starting the ``generate_atlas_lightcurves`` function')

    # SELECT SOURCES THAT NEED THEIR ATLAS FP LIGHTCURVES CREATED/UPDATED
    sqlQuery = u"""
        SELECT
                t.transientBucketId
            FROM
                transientBucket t ,pesstoObjects p
            WHERE
                p.transientBucketId=t.transientBucketId
                and t.survey = 'ATLAS FP' and t.limitingMag = 0
                and ((p.atlas_fp_lightcurve < t.dateCreated) or p.atlas_fp_lightcurve is null)
            GROUP BY t.transientBucketId;
    """
    rows = readquery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )
    transientIds = [r["transientBucketId"] for r in rows]

    total = len(transientIds)
    if total > 1000:
        print("ATLAS lightcurves need generated for %(total)s sources - generating next 1000" % locals())
        transientIds = transientIds[:100]
        total = len(transientIds)
    else:
        print("Generating ATLAS lightcurves for %(total)s sources" % locals())

    # SETUP THE INITIAL FIGURE FOR THE PLOT (ONLY ONCE)
    fig = plt.figure(
        num=None,
        figsize=(10, 10),
        dpi=100,
        facecolor=None,
        edgecolor=None,
        frameon=True)
    mpl.rc('ytick', labelsize=18)
    mpl.rc('xtick', labelsize=18)
    mpl.rcParams.update({'font.size': 22})

    # FORMAT THE AXES
    ax = fig.add_axes(
        [0.1, 0.1, 0.8, 0.8],
        polar=False,
        frameon=True)
    ax.set_xlabel('MJD', labelpad=20)
    ax.set_yticks([2.2])

    # RHS AXIS TICKS
    plt.setp(ax.xaxis.get_majorticklabels(),
             rotation=45, horizontalalignment='right')
    ax.xaxis.set_major_formatter(mtick.FormatStrFormatter('%5.0f'))

    y_formatter = mpl.ticker.FormatStrFormatter("%2.1f")
    ax.yaxis.set_major_formatter(y_formatter)
    ax.xaxis.grid(False)

    # ADD SECOND Y-AXIS
    ax2 = ax.twinx()
    ax2.yaxis.set_major_formatter(y_formatter)
    ax2.set_ylabel('Flux ($\mu$Jy)', rotation=-90.,  labelpad=27)
    ax2.grid(False)

    # ADD SECOND X-AXIS
    ax3 = ax.twiny()
    ax3.grid(True)
    plt.setp(ax3.xaxis.get_majorticklabels(),
             rotation=45, horizontalalignment='left')

    # CONVERTER TO CONVERT MJD TO DATE
    converter = conversions(
        log=log
    )

    if len(transientIds) < 3:
        plotPaths = []
        for transientBucketId in transientIds:
            plotPaths.append(
                plot_single_result(
                    log=log,
                    transientBucketId=transientBucketId,
                    fig=fig,
                    converter=converter,
                    ax=ax,
                    settings=settings)
            )
    else:
        log.info("""starting multiprocessing""")
        plotPaths = fmultiprocess(
            log=log,
            function=plot_single_result,
            inputArray=transientIds,
            poolSize=False,
            timeout=7200,
            fig=fig,
            converter=converter,
            ax=ax,
            settings=settings
        )
        log.info("""finished multiprocessing""")

    # UPDATE THE atlas_fp_lightcurve DATE FOR TRANSIENTS WE HAVE JUST
    # GENERATED PLOTS FOR
    if len(transientIds):
        transientIds = (",").join([str(t) for t in transientIds])
        sqlQuery = f"""update pesstoObjects set atlas_fp_lightcurve = NOW() where transientBucketID in ({transientIds})"""
        writequery(
            log=log,
            sqlQuery=sqlQuery,
            dbConn=dbConn
        )

    log.debug('completed the ``generate_atlas_lightcurves`` function')
    return None


def plot_single_result(
        transientBucketId,
        log,
        fig,
        converter,
        ax,
        settings):
    """*plot single result*

    **Key Arguments:**
        - `transientBucketId` -- thr transient ID for the source in the database
        - `log` -- logger
        - `fig` -- the matplotlib figure to use for the plot
        - `converter` -- converter to switch mjd to ut-date
        - `ax` -- plot axis
        - `settings` -- dictionary of settings (from yaml settings file)

    **Return:**
        - `filePath` -- path to the output PNG plot
    """
    log.info('starting the ``plot_single_result`` method')

    # SETUP DATABASE CONNECTION
    dbConn = database(
        log=log,
        dbSettings=settings["database settings"]
    ).connect()

    # GET THE DATA FROM THE DATABASE
    sqlQuery = u"""
        SELECT
            atlas_designation,
            mjd_obs,
            filter,
            marshall_mag as mag,
            marshall_mag_error as dm,
            fnu*1e29 as uJy,
            fnu_error*1e29 as duJy,
            snr,
            zp,
            marshall_limiting_mag as limiting_mag
        FROM
            fs_atlas_forced_phot
        WHERE
            (skyfit > 0) and
            atlas_designation in (SELECT distinct name
        FROM
            transientBucket
        WHERE
            survey = 'ATLAS FP'
                AND transientBucketId = %(transientBucketId)s
                AND dateDeleted IS NULL)
        and fnu is not null;
    """ % locals()
    epochs = readquery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn
    )

    ax2 = get_twin_axis(ax, "x")
    ax3 = get_twin_axis(ax, "y")

    # ax = fig.gca()
    epochs = sigma_clip_data(log=log, fpData=epochs)

    # c = cyan, o = arange
    magnitudes = {
        'c': {'mjds': [], 'mags': [], 'magErrs': []},
        'o': {'mjds': [], 'mags': [], 'magErrs': []},
        'I': {'mjds': [], 'mags': [], 'magErrs': []},
    }

    # SPLIT BY FILTER
    for epoch in epochs:
        if epoch["filter"] in ["c", "o", "I"]:
            magnitudes[epoch["filter"]]["mjds"].append(epoch["mjd_obs"])
            magnitudes[epoch["filter"]]["mags"].append(epoch["uJy"])
            magnitudes[epoch["filter"]]["magErrs"].append(epoch["duJy"])

    # STACK PHOTOMETRY IF REQUIRED
    magnitudes = stack_photometry(log=log,
                                  magnitudes=magnitudes, binningDays=1.)

    # ADD MAGNITUDES AND LIMITS FOR EACH FILTER
    handles = []

    # SET AXIS LIMITS FOR MAGNTIUDES
    upperMag = -99999999999
    lowerMag = 99999999999

    # DETERMINE THE TIME-RANGE OF DETECTION FOR THE SOURCE
    mjdList = magnitudes['o']['mjds'] + \
        magnitudes['c']['mjds'] + magnitudes['I']['mjds']
    if len(mjdList) == 0:
        log.error(f'{transientBucketId} does not have enough data to plot LC')
        return None
    lowerDetectionMjd = min(mjdList)
    upperDetectionMjd = max(mjdList)

    # DETERMIN MAGNITUDE RANGE
    allMags = magnitudes['o']['mags'] + magnitudes['c']['mags']
    magRange = max(allMags) - min(allMags)
    deltaMag = magRange * 0.1

    if len(magnitudes['o']['mjds']):
        orangeMag = ax.errorbar(magnitudes['o']['mjds'], magnitudes['o']['mags'], yerr=magnitudes[
            'o']['magErrs'], color='#FFA500', fmt='o', mfc='#FFA500', mec='#FFA500', zorder=1, ms=12., alpha=0.8, linewidth=1.2,  label='o-band mag ', capsize=10)

        # ERROBAR CAP THICKNESS
        orangeMag[1][0].set_markeredgewidth('0.7')
        orangeMag[1][1].set_markeredgewidth('0.7')
        handles.append(orangeMag)
        errMask = np.array(magnitudes['o']['magErrs'])
        np.putmask(errMask, errMask > 30, 30)

        if max(np.array(magnitudes['o']['mags']) + errMask) > upperMag:
            upperMag = max(
                np.array(magnitudes['o']['mags']) + errMask)
            upperMagIndex = np.argmax((
                magnitudes['o']['mags']) + errMask)

        if min(np.array(magnitudes['o']['mags']) - errMask) < lowerMag:
            lowerMag = min(
                np.array(magnitudes['o']['mags']) - errMask)
            lowerMagIndex = np.argmin((
                magnitudes['o']['mags']) - errMask)

    if len(magnitudes['c']['mjds']):
        cyanMag = ax.errorbar(magnitudes['c']['mjds'], magnitudes['c']['mags'], yerr=magnitudes[
            'c']['magErrs'], color='#2aa198', fmt='o', mfc='#2aa198', mec='#2aa198', zorder=1, ms=12., alpha=0.8, linewidth=1.2, label='c-band mag ', capsize=10)
        # ERROBAR CAP THICKNESS
        cyanMag[1][0].set_markeredgewidth('0.7')
        cyanMag[1][1].set_markeredgewidth('0.7')
        handles.append(cyanMag)
        errMask = np.array(magnitudes['c']['magErrs'])
        np.putmask(errMask, errMask > 30, 30)

        if max(np.array(magnitudes['c']['mags']) + errMask) > upperMag:
            upperMag = max(
                np.array(magnitudes['c']['mags']) + errMask)
            upperMagIndex = np.argmax((
                magnitudes['c']['mags']) + errMask)

        if min(np.array(magnitudes['c']['mags']) - errMask) < lowerMag:
            lowerMag = min(
                np.array(magnitudes['c']['mags']) - errMask)
            lowerMagIndex = np.argmin(
                (magnitudes['c']['mags']) - errMask)

    # if self.firstPlot:
    plt.legend(handles=handles, prop={
        'size': 13.5}, bbox_to_anchor=(0.95, 1.2), loc=0, borderaxespad=0., ncol=4, scatterpoints=1)

    # SET THE TEMPORAL X-RANGE
    allMjd = magnitudes['o']['mjds'] + magnitudes['c']['mjds']
    xmin = min(allMjd) - 5.
    xmax = max(allMjd) + 5.
    mjdRange = xmax - xmin
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([lowerMag - deltaMag, upperMag + deltaMag])

    # PLOT THE MAGNITUDE SCALE
    axisUpperFlux = upperMag
    axisLowerFlux = 1e-29
    axisLowerMag = -2.5 * math.log10(axisLowerFlux) + 23.9
    if axisUpperFlux > 0.:
        axisUpperMag = -2.5 * math.log10(axisUpperFlux) + 23.9
    else:
        axisUpperMag = None
    if axisUpperMag:
        ax.set_ylabel('Apparent Magnitude', labelpad=15)
        magLabels = [20., 17.0, 15.0, 14.0, 13.5, 13.0]
        if axisUpperMag < 14:
            magLabels = [20.0, 16.0, 15.0, 14.0,
                         13.0, 12.5, 12.0, 11.5, 11.0]
        elif axisUpperMag < 17:
            magLabels = [20.,
                         18.0, 17.0, 16.5, 16.0, 15.5, 15.0]
        elif axisUpperMag < 18:
            magLabels = [20., 19.5, 19.0, 18.5,
                         18.0, 17.5, 17.0, 16.5, 16.0, 15.5, 15.0]
        elif axisUpperMag < 20:
            magLabels = [20.5, 20.0, 19.5, 19.0, 18.5, 18.0]

        magFluxes = [pow(10, old_div(-(m - 23.9), 2.5)) for m in magLabels]

        ax.yaxis.set_major_locator(mtick.FixedLocator((magFluxes)))
        ax.yaxis.set_major_formatter(mtick.FixedFormatter((magLabels)))
    else:
        ax.set_yticks([])

    # ADD SECOND Y-AXIS
    ax2.set_ylim([lowerMag - deltaMag, upperMag + deltaMag])

    # RELATIVE TIME SINCE DISCOVERY
    lower, upper = ax.get_xlim()
    utLower = converter.mjd_to_ut_datetime(mjd=lower, datetimeObject=True)
    utUpper = converter.mjd_to_ut_datetime(mjd=upper, datetimeObject=True)
    ax3.set_xlim([utLower, utUpper])

    if mjdRange > 365:
        ax3.xaxis.set_major_formatter(dates.DateFormatter('%b %d %y'))
    else:
        ax3.xaxis.set_major_formatter(dates.DateFormatter('%b %d'))

    # FIND THE CACHE DIR FOR THE SOURCE
    cacheDirectory = settings[
        "cache-directory"] + "/transients/" + str(transientBucketId)
    if not os.path.exists(cacheDirectory):
        os.makedirs(cacheDirectory)
    filePath = cacheDirectory + "/atlas_fp_lightcurve.png"
    plt.savefig(filePath, bbox_inches='tight', transparent=False,
                pad_inches=0.1, optimize=True, progressive=True)

    try:
        cyanMag.remove()
    except:
        pass

    try:
        orangeMag.remove()
    except:
        pass

    firstPlot = False

    log.debug('completed the ``plot_single_result`` method')
    return filePath


def sigma_clip_data(
        log,
        fpData,
        clippingSigma=2.2):
    """*clean up rouge data from the files by performing some basic clipping*

    **Key Arguments:**

    - `fpData` -- data dictionary of force photometry
    - `clippingSigma` -- the level at which to clip flux data

    **Return:**

    - `epochs` -- sigma clipped and cleaned epoch data
    """
    log.info('starting the ``sigma_clip_data`` function')

    mjdMin = False
    mjdMax = False

    # PARSE DATA WITH SOME FIXED CLIPPING
    oepochs = []
    cepochs = []

    for row in fpData:
        # REMOVE VERY HIGH ERROR DATA POINTS
        if row["duJy"] > 4000:
            continue
        if mjdMin and mjdMax:
            if row["mjd_obs"] < mjdMin or row["mjd_obs"] > mjdMax:
                continue
        if row["filter"] == "c":
            cepochs.append(row)
        if row["filter"] == "o":
            oepochs.append(row)

    # SORT BY MJD
    cepochs = sorted(cepochs, key=itemgetter('mjd_obs'), reverse=False)
    oepochs = sorted(oepochs, key=itemgetter('mjd_obs'), reverse=False)

    # SIGMA-CLIP THE DATA WITH A ROLLING WINDOW
    cdataFlux = []
    cdataFlux[:] = [row["uJy"] for row in cepochs]
    odataFlux = []
    odataFlux[:] = [row["uJy"] for row in oepochs]

    maskList = []
    for flux in [cdataFlux, odataFlux]:
        fullMask = rolling_window_sigma_clip(
            log=log,
            array=flux,
            clippingSigma=clippingSigma,
            windowSize=11)
        maskList.append(fullMask)

    try:
        cepochs = [e for e, m in zip(
            cepochs, maskList[0]) if m == False]
    except:
        cepochs = []

    try:
        oepochs = [e for e, m in zip(
            oepochs, maskList[1]) if m == False]
    except:
        oepochs = []

    log.debug('completed the ``sigma_clip_data`` function')
    return cepochs + oepochs


def stack_photometry(
        log,
        magnitudes,
        binningDays=1.):
    """*stack the photometry for the given temporal range*

    **Key Arguments:**
        - `magnitudes` -- dictionary of photometry divided into filter sets
        - `binningDays` -- the binning to use (in days)

    **Return:**
        - `summedMagnitudes` -- the stacked photometry
    """
    log.debug('starting the ``stack_photometry`` method')

    # IF WE WANT TO 'STACK' THE PHOTOMETRY
    summedMagnitudes = {
        'c': {'mjds': [], 'mags': [], 'magErrs': [], 'n': []},
        'o': {'mjds': [], 'mags': [], 'magErrs': [], 'n': []},
        'I': {'mjds': [], 'mags': [], 'magErrs': [], 'n': []},
    }

    # MAGNITUDES/FLUXES ARE DIVIDED IN UNIQUE FILTER SETS - SO ITERATE OVER
    # FILTERS
    allData = []
    for fil, data in list(magnitudes.items()):
        # WE'RE GOING TO CREATE FURTHER SUBSETS FOR EACH UNQIUE MJD (FLOORED TO AN INTEGER)
        # MAG VARIABLE == FLUX (JUST TO CONFUSE YOU)
        distinctMjds = {}
        for mjd, flx, err in zip(data["mjds"], data["mags"], data["magErrs"]):
            # DICT KEY IS THE UNIQUE INTEGER MJD
            key = str(int(math.floor(mjd / float(binningDays))))
            # FIRST DATA POINT OF THE NIGHTS? CREATE NEW DATA SET
            if key not in distinctMjds:
                distinctMjds[key] = {
                    "mjds": [mjd],
                    "mags": [flx],
                    "magErrs": [err]
                }
            # OR NOT THE FIRST? APPEND TO ALREADY CREATED LIST
            else:
                distinctMjds[key]["mjds"].append(mjd)
                distinctMjds[key]["mags"].append(flx)
                distinctMjds[key]["magErrs"].append(err)

        # ALL DATA NOW IN MJD SUBSETS. SO FOR EACH SUBSET (I.E. INDIVIDUAL
        # NIGHTS) ...
        for k, v in list(distinctMjds.items()):
            # GIVE ME THE MEAN MJD
            meanMjd = old_div(sum(v["mjds"]), len(v["mjds"]))
            summedMagnitudes[fil]["mjds"].append(meanMjd)
            # GIVE ME THE MEAN FLUX
            meanFLux = old_div(sum(v["mags"]), len(v["mags"]))
            summedMagnitudes[fil]["mags"].append(meanFLux)
            # GIVE ME THE COMBINED ERROR
            combError = sum(v["magErrs"]) / len(v["magErrs"]
                                                ) / math.sqrt(len(v["magErrs"]))
            summedMagnitudes[fil]["magErrs"].append(combError)
            # GIVE ME NUMBER OF DATA POINTS COMBINED
            n = len(v["mjds"])
            summedMagnitudes[fil]["n"].append(n)
            allData.append({
                'MJD': f'{meanMjd:0.2f}',
                'uJy': f'{meanFLux:0.2f}',
                'duJy': f'{combError:0.2f}',
                'F': fil,
                'n': n
            })

    log.debug('completed the ``stack_photometry`` method')
    return summedMagnitudes


def get_twin(ax, axis):

    for sibling in siblings:
        if sibling.bbox.bounds == ax.bbox.bounds and sibling is not ax:
            return sibling
    return None


def get_twin_axis(ax, axis):
    assert axis in ("x", "y")
    for other_ax in ax.figure.axes:
        if other_ax is ax:
            siblings = getattr(ax, f"get_shared_{axis}_axes")().get_siblings(ax)
            for sibling in siblings:
                if sibling.bbox.bounds == ax.bbox.bounds and sibling is not ax:
                    return sibling
    return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*Generate the standard multi-survey lightcurve plots for the marshall transients*

:Author:
    David Young
"""
from __future__ import print_function
from builtins import str
from builtins import zip
from builtins import range
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals import fmultiprocess
from fundamentals.mysql import readquery, writequery
from fundamentals.mysql import database
from astrocalc.times import conversions, now
import numpy as np
# SUPPRESS MATPLOTLIB WARNINGS
import warnings
warnings.filterwarnings("ignore")
import matplotlib as mpl
from matplotlib import dates
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.ticker as mtick
from matplotlib.backends.backend_pdf import PdfPages
import math

from numpy.polynomial.chebyshev import chebfit, chebval


class marshall_lightcurves(object):
    """
    *The worker class for the marshall_lightcurves module*

    **Key Arguments**

    - ``log`` -- logger
    - ``settings`` -- the settings dictionary
    - ``dbConn`` -- the database connection for the mrshall
    - ``transientBucketIds`` -- the transientBucketId(s) requiring lightcurves to be regenerated. (int or list)


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To initiate a marshall_lightcurves object, use the following:

    ```python
    from marshallEngine.lightcurves import marshall_lightcurves
    lc = marshall_lightcurves(
        log=log,
        dbConn=dbConn,
        settings=settings,
        transientBucketIds=[28421489, 28121353, 4637952, 27409808]
    )
    lc.plot()
    ```

    """

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
            transientBucketIds=[]
    ):
        self.log = log
        log.debug("instansiating a new 'marshall_lightcurves' object")
        self.settings = settings
        self.dbConn = dbConn
        self.transientBucketIds = transientBucketIds
        # xt-self-arg-tmpx

        # CONVERT TRANSIENTBUCKETIDS TO LIST
        if not isinstance(self.transientBucketIds, list):
            self.transientBucketIds = [self.transientBucketIds]

        return None

    def _select_data_for_transient(
            self,
            transientBucketId):
        """*get the transient lightcurve data from the marshall database*

        **Key Arguments**

        - ``transientBucketId`` -- the transientBucketId of source to get data for

        """
        self.log.debug('starting the ``_select_data_for_transient`` method')

        # SELECT THE DATA REQUIRED FOR THIS LIGHTCURVE PLOT
        sqlQuery = """
            SELECT
                transientBucketId,
                survey,
                magnitude,
                magnitudeError,
                observationDate,
                observationMJD,
                IFNULL(filter, '?') as filter,
                name,
                limitingMag,
                concat(ROUND(observationMJD, 2), SUBSTRING(filter, 1, 1)) as uniqueConstraint,
                concat(ROUND(observationMJD, 0), SUBSTRING(filter, 1, 1)) as limitConstraint
            FROM
                transientBucket
            WHERE
                replacedByRowId = 0
                    AND transientBucketId = %(transientBucketId)s
                    AND magnitude IS NOT NULL
                    -- AND filter is not null
            ORDER BY uniqueConstraint ASC, magnitudeError desc, magnitude asc
        """ % locals()
        transientData = readquery(
            sqlQuery=sqlQuery,
            dbConn=self.dbConn,
            log=self.log
        )

        theseMags = []
        theseMags[:] = [a
                        for a in transientData if a["limitingMag"] == 0]

        # CLIP OUTLIERS
        if len(theseMags) > 4:
            allmags = []
            allmags[:] = [a["magnitude"] for a in theseMags]
            mean = np.mean(allmags)
            std = np.std(allmags)

            if std > 0.1:
                allmags[:] = [a for a in transientData if (abs(
                    a["magnitude"] - mean) < 2 * std) or a["limitingMag"] == 1]
                transientData = allmags

        # NO DATA?
        if len(transientData) == 0:
            return False, False, False

        # SPLIT DATA BY FILTER - MAGNITUDES AND LIMITS
        filterList = []
        filterList[:] = set([r["filter"][0] for r in transientData])

        dataset = {}
        flatdata = {"mag": [], "mjd": []}
        flatLimits = {"mag": [], "mjd": []}
        for f in filterList:
            mag = []
            magErr = []
            magMjd = []
            limit = []
            limitMjd = []
            magNoErr = []
            magNoErrMjd = []
            magNoErrFudge = []
            catch = []
            limitCatcher = {}
            for r in transientData:
                if r["filter"][0] == f and r["uniqueConstraint"] not in catch:
                    if r["limitingMag"] == 0 and r["magnitudeError"]:
                        mag.append(r["magnitude"])
                        magErr.append(r["magnitudeError"])
                        magMjd.append(r["observationMJD"])
                        flatdata["mag"].append(r["magnitude"])
                        flatdata["mjd"].append(r["observationMJD"])
                        catch.append(r["uniqueConstraint"])
                    elif r["limitingMag"] == 0 and not r["magnitudeError"]:
                        magNoErr.append(r["magnitude"])
                        magNoErrFudge.append(0.3)
                        magNoErrMjd.append(r["observationMJD"])
                        flatdata["mag"].append(r["magnitude"])
                        flatdata["mjd"].append(r["observationMJD"])
                        catch.append(r["uniqueConstraint"])
                    elif r["limitConstraint"] not in limitCatcher:
                        limitCatcher[r["limitConstraint"]] = [
                            r["magnitude"], r["observationMJD"]]
                    elif limitCatcher[r["limitConstraint"]][0] < r["magnitude"]:
                        limitCatcher[r["limitConstraint"]] = [
                            r["magnitude"], r["observationMJD"]]

            for k, v in list(limitCatcher.items()):
                limit.append(v[0])
                limitMjd.append(v[1])
                flatLimits["mag"].append(v[0])
                flatLimits["mjd"].append(v[1])

            dataset[f] = {
                "limit": limit,
                "limitMjd": limitMjd,
                "mag": mag,
                "magErr": magErr,
                "magMjd": magMjd,
                "magNoErr": magNoErr,
                "magNoErrFudge": magNoErrFudge,
                "magNoErrMjd": magNoErrMjd
            }

        if len(flatdata["mag"]) == 0:
            return False, False, False

        self.log.debug('completed the ``_select_data_for_transient`` method')
        return dataset, flatdata, flatLimits

    def _create_lightcurve_plot_file(
            self,
            dataset,
            flatdata,
            flatLimits,
            objectNames,
            saveLocation,
            saveFileName):
        """*Generate the lightcurve and save to file*

        **Key Arguments**

        - ``log`` -- logger
        - ``dataset`` -- the observational dataset split into filters (and then mags, limits etc)
        - ``flatdata`` -- a flattened dataset to determine current magnitude
        - ``flatLimits`` -- a flattened dataset of non-detection limits
        - ``objectNames`` -- a single name or a list of names
        - ``saveLocation`` -- the folder to save the plot file to
        - ``saveFileName`` -- the filename to give the plot file (without extension)


        **Return**

        - ``filepath`` -- path to the lightcurve file
        - ``currentMag`` -- a prediction of the current magnitude if there is enough recent data
        - ``gradient`` -- a prediction of the gradient of recent data (on rise or decline?)

        """
        self.log.debug('starting the ``_create_lightcurve_plot_file`` method')

        # CONVERTER TO CONVERT MJD TO DATE
        converter = conversions(
            log=self.log
        )

        # INITIATE THE PLOT FIGURE - SQUARE
        fig = plt.figure(
            num=None,
            figsize=(10, 10),
            dpi=100,
            facecolor=None,
            edgecolor=None,
            frameon=True)
        ax = fig.add_subplot(1, 1, 1)

        # TICK LABEL SIZE
        mpl.rc('ytick', labelsize=25)
        mpl.rc('xtick', labelsize=25)
        mpl.rcParams.update({'font.size': 25})

        # INITIAL RESTRICTIONS
        currentMag = -9999
        gradient = -9999

        # WHAT IS TODAY MJD (FIR CURRENT MAG ESTIMATE)
        todayMjd = now(
            log=self.log
        ).get_mjd()

        # MAKE ARRAYS OF TIME AND MAG FOR PLOTS
        bigTimeArray, bigMagArray = np.array(
            flatdata["mjd"]), np.array(flatdata["mag"])
        # SORT TWO LIST BASED ON FIRST
        bigTimeArray, bigMagArray = zip(
            *[(x, y) for x, y in sorted(zip(bigTimeArray, bigMagArray))])

        # BIN DATA FOR POLYNOMIALS
        binData = True
        if binData is True:
            distinctMjds = {}
            for mjd, mag in zip(bigTimeArray, bigMagArray):
                # DICT KEY IS THE UNIQUE INTEGER MJD
                key = str(int(math.floor(mjd / 1.0)))
                # FIRST DATA POINT OF THE NIGHTS? CREATE NEW DATA SET
                if key not in distinctMjds:
                    distinctMjds[key] = {
                        "mjds": [mjd],
                        "mags": [mag]
                    }
                # OR NOT THE FIRST? APPEND TO ALREADY CREATED LIST
                else:
                    distinctMjds[key]["mjds"].append(mjd)
                    distinctMjds[key]["mags"].append(mag)

            # ALL DATA NOW IN MJD SUBSETS. SO FOR EACH SUBSET (I.E. INDIVIDUAL
            # NIGHTS) ...
            summedMagnitudes = {
                'mjds': [],
                'mags': []
            }
            for k, v in list(distinctMjds.items()):
                # GIVE ME THE MEAN MJD
                meanMjd = sum(v["mjds"]) / len(v["mjds"])
                summedMagnitudes["mjds"].append(meanMjd)
                # GIVE ME THE MEAN MAG
                meanMag = sum(v["mags"]) / len(v["mags"])
                summedMagnitudes["mags"].append(meanMag)

            bigTimeArray = summedMagnitudes["mjds"]
            bigMagArray = summedMagnitudes["mags"]

        bigTimeArray = np.array(bigTimeArray)
        bigMagArray = np.array(bigMagArray)

        # DETERMINE SENSIBLE AXIS LIMITS FROM FLATTENED DATA
        # LIMITS HERE ARE LOWER AND UPPER MJDS FOR X-AXIS
        xLowerLimit = bigTimeArray.min()
        xUpperLimit = bigTimeArray.max()
        latestTime = xUpperLimit
        xBorder = math.fabs((xUpperLimit - xLowerLimit)) * 0.1
        if xBorder < 5:
            xBorder = 5.
        xLowerLimit -= xBorder
        xUpperLimit += xBorder
        fixedXUpperLimit = xUpperLimit
        timeRange = xUpperLimit - xLowerLimit

        # POLYNOMIAL CONSTAINTS USING COMBINED DATASETS
        # POLYNOMIAL/LINEAR SETTINGS
        # SETTINGS FILE
        polyOrder = 5
        # EITHER USE DATA IN THESE LAST NUMBER OF DAYS OR ...
        lastNumDays = 10.
        # ... IF NOT ENOUGH DATA USE THE LAST NUMBER OF DATA POINTS
        predictCurrentMag = True
        lastNumDataPoints = 3
        numAnchors = 3
        anchorSeparation = 70
        latestMag = bigMagArray[0]
        anchorPointMag = latestMag + 0.5
        polyTimeArray, polyMagArray = [], []

        # QUIT IF NOT ENOUGH DATA FOR POLYNOMIAL
        if len(bigTimeArray) <= lastNumDataPoints or timeRange < 3.:
            predictCurrentMag = False

        if max(bigTimeArray) < todayMjd - 120:
            predictCurrentMag = False

        if predictCurrentMag:
            # USE ONLY THE LAST N DAYS OF DATA FOR LINEAR FIT
            mask = np.where(bigTimeArray -
                            bigTimeArray.max() < -lastNumDays, False, True)

            # DETERMINE GRADIENT OF SLOPE FROM LAST `LASTNUMDAYS` DAYS
            linearTimeArray = bigTimeArray[mask]
            linearMagArray = bigMagArray[mask]
            # FIT AND PLOT THE POLYNOMIAL ASSOCSIATED WITH ALL DATA SETS
            thisLinear = chebfit(linearTimeArray, linearMagArray, 1)
            gradient = thisLinear[1]

            firstAnchorPointTime = anchorSeparation + latestTime

            # CREATE THE ARRAY OF DATA USED TO GERNERATE THE POLYNOMIAL
            polyTimeArray = bigTimeArray
            polyMagArray = bigMagArray

            # ANCHOR THE POLYNOMIAL IN THE FUTURE SO THAT ALL PREDICTED LIGHTCURVES
            # EVENTUALLY FADE TO NOTHING
            extraTimes = np.arange(0, numAnchors) * \
                anchorSeparation + firstAnchorPointTime
            extraMags = np.ones(numAnchors) * anchorPointMag
            polyTimeArray = np.append(polyTimeArray, extraTimes)
            polyMagArray = np.append(polyMagArray, extraMags)

            # POLYNOMIAL LIMTIS
            xPolyLowerLimit = min(polyTimeArray) - 2.0
            xPolyUpperLimit = max(polyTimeArray) + 2.0

        # SET AXIS LIMITS
        xUpperLimit = 5
        yLowerLimit = min(bigMagArray) - 0.3
        yUpperLimit = max(bigMagArray) + 0.5
        yBorder = math.fabs((yUpperLimit - yLowerLimit)) * 0.1
        yLowerLimit -= yBorder
        yUpperLimit += yBorder

        # EXTEND LOWER X-LIMIT FOR NON-DETECTIONS
        xLowerTmp = xLowerLimit
        for t, m in zip(flatLimits["mjd"], flatLimits["mag"]):
            if m > yLowerLimit and t < xLowerTmp + 2 and t > xLowerLimit - 40:
                xLowerTmp = t - 2
        xLowerLimit = xLowerTmp

        if predictCurrentMag:
            thisPoly = chebfit(polyTimeArray, polyMagArray, polyOrder)
            # FLATTEN INTO A FUNCTION TO MAKE PLOTTING EASIER
            xData = np.arange(xPolyLowerLimit, todayMjd + 50, 1)
            flatLinear = chebval(xData, thisLinear)
            flatPoly = chebval(xData, thisPoly)
            plt.plot(xData, flatPoly, label="poly")
            plt.plot(xData, flatLinear, label="linear")

            # PREDICT A CURRENT MAGNITUDE FROM THE PLOT

            currentMag = chebval(todayMjd, thisPoly)
            self.log.debug(
                'currentMag: %(currentMag)0.2f, m=%(gradient)s' % locals())

            ls = "*g" % locals()
            currentMagArray = np.array([currentMag])
            nowArray = np.array([todayMjd])
            line = ax.plot(nowArray, currentMagArray,
                           ls, label="current estimate")

            lineExtras = ax.plot(extraTimes, extraMags, "+")

            # SET THE AXES / VIEWPORT FOR THE PLOT
            # if currentMag < yLowerLimit:
            #     yLowerLimit = currentMag - 0.4

        if currentMag > 23:
            currentMag = -9999.

        plt.clf()
        plt.cla()
        ax = fig.add_subplot(1, 1, 1)
        # print(currentMag)
        # print(bigTimeArray)
        # print(bigMagArray)

        # PLOT DATA VIA FILTER. MAGS AND LIMITS
        filterColor = {
            "r": "#29a329",
            "g": "#268bd2",
            "G": "#859900",
            "o": "#cb4b16",
            "c": "#2aa198",
            "U": "#6c71c4",
            "B": "blue",
            "V": "#008000",
            "R": "#e67300",
            "I": "#dc322f",
            "w": "#cc2900",
            "y": "#ff6666",
            "z": "#990000",
        }
        i = 0
        handles = []
        handlesAdded = []
        for k, v in list(dataset.items()):
            mag = v["mag"]
            magErr = v["magErr"]
            magMjd = v["magMjd"]
            limit = v["limit"]
            limitMjd = v["limitMjd"]
            magNoErr = v["magNoErr"]
            magNoErrMjd = v["magNoErrMjd"]
            magNoErrFudge = v["magNoErrFudge"]

            if k in filterColor:
                color = filterColor[k]
            else:
                color = "black"

            if len(limit):
                for l, m in zip(limit, limitMjd):
                    plt.text(m, l, u"\u21A7", fontname='STIXGeneral',
                             size=30, va='top', ha='center', clip_on=True, color=color, zorder=1)
            if len(magNoErr):
                theseMags = ax.errorbar(magNoErrMjd, magNoErr, yerr=magNoErrFudge, color=color, fmt='o', mfc=color,
                                        mec=color, zorder=2, ms=12., alpha=0.8, linewidth=1.2,  label=k, capsize=0)
                theseMags[-1][0].set_linestyle('--')

            if len(mag):
                theseMags = ax.errorbar(magMjd, mag, yerr=magErr, color=color, fmt='o', mfc=color,
                                        mec=color, zorder=3, ms=12., alpha=0.8, linewidth=1.2,  label=k, capsize=10)

            if not len(mag):
                theseMags = ax.errorbar([-500], [20], yerr=[0.2], color=color, fmt='o', mfc=color,
                                        mec=color, zorder=3, ms=12., alpha=0.8, linewidth=1.2,  label=k, capsize=10)

            if k not in handlesAdded:
                handles.append(theseMags)
                handlesAdded.append(k)

        # ADD LEGEND
        plt.legend(handles=handles, prop={
                   'size': 13.5}, bbox_to_anchor=(1., 1.25), loc=0, borderaxespad=0., ncol=18, scatterpoints=1)

        # RHS AXIS TICKS
        plt.setp(ax.xaxis.get_majorticklabels(),
                 rotation=45, horizontalalignment='right')
        ax.xaxis.set_major_formatter(mtick.FormatStrFormatter('%5.0f'))

        # CHANGE PLOT TO FIXED TIME
        # SETUP THE AXES
        xUpperLimit = fixedXUpperLimit
        ax.set_xlabel('MJD',  labelpad=20, fontsize=30)
        ax.set_ylabel('Magnitude',  labelpad=20, fontsize=30)
        ax.set_title('')
        ax.set_xlim([xLowerLimit, xUpperLimit])
        ax.set_ylim([yUpperLimit, yLowerLimit])
        ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%d'))

        # GENERATE UT DATE AXIS FOR TOP OF PLOT
        lower, upper = ax.get_xlim()
        utLower = converter.mjd_to_ut_datetime(mjd=lower, datetimeObject=True)
        utUpper = converter.mjd_to_ut_datetime(mjd=upper, datetimeObject=True)
        ax3 = ax.twiny()
        ax3.set_xlim([utLower, utUpper])
        ax3.grid(True)
        ax.xaxis.grid(False)
        plt.setp(ax3.xaxis.get_majorticklabels(),
                 rotation=45, horizontalalignment='left', fontsize=14)
        ax3.xaxis.set_major_formatter(dates.DateFormatter('%b %d, %y'))

        # Y TICK FORMAT
        y_formatter = mpl.ticker.FormatStrFormatter("%2.1f")
        ax.yaxis.set_major_formatter(y_formatter)

        # PRINT CURRENT MAG AS SANITY CHECK
        # fig.text(0.1, 1.02, currentMag, ha="left", fontsize=40)

        # RECURSIVELY CREATE MISSING DIRECTORIES
        if not os.path.exists(saveLocation):
            try:
                os.makedirs(saveLocation)
            except:
                pass
        # SAVE THE PLOT
        filepath = """%(saveLocation)s%(saveFileName)s.png""" % locals()
        plt.savefig(filepath, format='PNG', bbox_inches='tight', transparent=False,
                    pad_inches=0.4)
        # plt.show()
        plt.clf()  # clear figure
        plt.close()

        # TEST THAT PLOT FILE HAS ACTUALLY BEEN GENERATED
        try:
            with open(filepath):
                pass
            fileExists = True
        except IOError:
            raise IOError(
                "the path --pathToFile-- %s does not exist on this machine" %
                (filepath,))
            filepath = False

        self.log.debug('completed the ``_create_lightcurve_plot_file`` method')

        return filepath, currentMag, gradient

    def plot(
            self):
        """*generate a batch of lightcurves using multiprocessing given their transientBucketIds*

        **Return**

        - ``filepath`` -- path to the last generated plot file


        **Usage**

        ```python
        from marshallEngine.lightcurves import marshall_lightcurves
        lc = marshall_lightcurves(
            log=log,
            dbConn=dbConn,
            settings=settings,
            transientBucketIds=[28421489, 28121353, 4637952, 27409808]
        )
        lc.plot()
        ```

        """
        self.log.debug('starting the ``plot`` method')

        # DEFINE AN INPUT ARRAY
        total = len(self.transientBucketIds)

        thisDict = {"database settings": self.settings["database settings"]}

        if total:
            print("updating lightcurves for %(total)s transients" % locals())
            print()

        # USE IF ISSUES IN _plot_one FUNCTION
        # for transientBucketId in self.transientBucketIds:
        #     _plot_one(
        #         transientBucketId=transientBucketId,
        #         log=self.log,
        #         settings=self.settings
        #     )

        results = fmultiprocess(log=self.log, function=_plot_one,
                                inputArray=self.transientBucketIds, poolSize=False, timeout=3600, settings=self.settings)

        sqlQuery = ""
        updatedTransientBucketIds = []
        for t, r in zip(self.transientBucketIds, results):
            if not r[0]:
                # LIGHTCURVE NOT GENERATED
                continue
            updatedTransientBucketIds.append(t)
            filepath = r[0]
            currentMagnitude = r[1]
            gradient = r[2]
            sqlQuery += """update transientBucketSummaries set currentMagnitudeEstimate = %(currentMagnitude)s, currentMagnitudeEstimateUpdated = NOW(), recentSlopeOfLightcurve = %(gradient)s where transientBucketId = %(t)s;
            """ % locals()
        ids = []
        ids[:] = [str(i) for i in updatedTransientBucketIds]
        updatedTransientBucketIds = (",").join(ids)
        sqlQuery += "update pesstoObjects set master_pessto_lightcurve = 1 where transientBucketId in (%(updatedTransientBucketIds)s);" % locals(
        )

        if len(updatedTransientBucketIds):
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn,
            )
        else:
            filepath = False

        self.log.debug('completed the ``plot`` method')

        return filepath


def _plot_one(
        transientBucketId,
        log,
        settings):
    """*plot a single transeint lightcurve*

    **Key Arguments**

    - ``transientBucketId`` -- the id of the single transient to plot.
    - ``settings`` -- dictionary of settings
    - ``dbConn`` -- marshall database connection


    **Return**

    - ``filepath`` -- path to the plot file
    - ``currentMag`` -- an estimate of the current magnitude (from slope of recent LC). -9999 if inaccurate.
    - ``gradient`` -- gradient of slope of the recent LC. -9999 if inaccurate.

    """
    log.debug('starting the ``_plot_one`` method')

    # MULTIPROCESSING NEEDS ONE CONNECTION PERPROCESS
    sys.stdout.write("\x1b[1A\x1b[2K")
    print("updating LC for transient %(transientBucketId)s" % locals())
    dbConn = database(
        log=log,
        dbSettings=settings["database settings"]
    ).connect()

    # LC OBJECT
    lc = marshall_lightcurves(
        log=log,
        dbConn=dbConn,
        settings=settings,
        transientBucketIds=transientBucketId
    )

    cacheFolder = settings[
        "cache-directory"] + "/transients/"
    saveLocation = """%(cacheFolder)s/%(transientBucketId)s/""" % locals()

    # SELECT DATA AND PLOT THE SOURCE
    dataset, flatdata, flatLimits = lc._select_data_for_transient(
        transientBucketId)
    if dataset:
        filepath, currentMag, gradient = lc._create_lightcurve_plot_file(
            dataset=dataset,
            flatdata=flatdata,
            flatLimits=flatLimits,
            objectNames="test object",
            saveLocation=saveLocation,
            saveFileName="master_lightcurve"
        )
    else:
        return False, False, False

    log.debug('completed the ``_plot_one`` method')
    return filepath, currentMag, gradient

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*import the panstarrs stream into the marshall*

:Author:
    David Young
"""
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
from fundamentals.mysql import writequery


class data(basedata):
    """
    *Import the PanSTARRS transient data into the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``dbConn`` -- the marshall database connection
    - ``settings`` -- the settings dictionary


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a data object, use the following:

    ```python
    from marshallEngine.feeders.panstarrs.data import data
    ingester = data(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).ingest(withinLastDays=withInLastDay)   
    ```

    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn

        self.fsTableName = "fs_panstarrs"
        self.survey = "Pan-STARRS"

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments**

        - ``withinLastDays`` -- within the last number of days. *Default: 50*

        """
        self.log.debug('starting the ``ingest`` method')

        allLists = []
        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["ps13pi"]["summary csv"],
            user=self.settings["credentials"]["ps13pi"]["username"],
            pwd=self.settings["credentials"]["ps13pi"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="ps13pi", withinLastDays=withinLastDays))

        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["ps13pi"]["recurrence csv"],
            user=self.settings["credentials"]["ps13pi"]["username"],
            pwd=self.settings["credentials"]["ps13pi"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="ps13pi", withinLastDays=withinLastDays))

        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["ps23pi"]["summary csv"],
            user=self.settings["credentials"]["ps23pi"]["username"],
            pwd=self.settings["credentials"]["ps23pi"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="ps23pi", withinLastDays=withinLastDays))

        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["ps23pi"]["recurrence csv"],
            user=self.settings["credentials"]["ps23pi"]["username"],
            pwd=self.settings["credentials"]["ps23pi"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="ps23pi", withinLastDays=withinLastDays))

        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["pso3"]["summary csv"],
            user=self.settings["credentials"]["pso3"]["username"],
            pwd=self.settings["credentials"]["pso3"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="pso3", withinLastDays=withinLastDays))
        csvDicts = self.get_csv_data(
            url=self.settings["panstarrs urls"]["pso3"]["recurrence csv"],
            user=self.settings["credentials"]["pso3"]["username"],
            pwd=self.settings["credentials"]["pso3"]["password"]
        )
        allLists.extend(self._clean_data_pre_ingest(
            surveyName="pso3", withinLastDays=withinLastDays))

        self.dictList = allLists
        self._import_to_feeder_survey_table()

        self.insert_into_transientBucket()

        # FIX ODD PANSTARRS COORDINATES
        sqlQuery = """update transientBucket set raDeg = raDeg+360.0 where raDeg  < 0;""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.debug('completed the ``ingest`` method')
        return None

    def _clean_data_pre_ingest(
            self,
            surveyName,
            withinLastDays=False):
        """*clean up the list of dictionaries containing the PS data, pre-ingest*

        **Key Arguments**

        - ``surveyName`` -- the PS survey name
        -  ``withinLastDays`` -- the lower limit of observations to include (within the last N days from now). Default *False*, i.e. no limit


        **Return**

        - ``dictList`` -- the cleaned list of dictionaries ready for ingest


        **Usage**

        To clean the data from the PS 3pi survey:

        ```python
        dictList = ingesters._clean_data_pre_ingest(surveyName="3pi")
        ```

        Note you will also be able to access the data via ``ingester.dictList``

        """
        self.log.debug('starting the ``_clean_data_pre_ingest`` method')

        self.dictList = []

        # CALC MJD LIMIT
        if withinLastDays:
            mjdLimit = now(
                log=self.log
            ).get_mjd() - float(withinLastDays)

        for row in self.csvDicts:
            # IF NOW IN THE LAST N DAYS - SKIP
            if withinLastDays and float(row["mjd_obs"]) < mjdLimit:
                continue
            if float(row["ra_psf"]) < 0:
                row["ra_psf"] = 360. + float(row["ra_psf"])
            thisDictionary = {}

            thisDictionary["candidateID"] = row["ps1_designation"]
            thisDictionary["ra_deg"] = row["ra_psf"]
            thisDictionary["dec_deg"] = row["dec_psf"]
            thisDictionary["mag"] = row["cal_psf_mag"]
            thisDictionary["magerr"] = row["psf_inst_mag_sig"]
            thisDictionary["observationMJD"] = row["mjd_obs"]
            thisDictionary["filter"] = row["filter"]

            try:
                thisDictionary["discDate"] = row["followup_flag_date"]
            except:
                pass
            thisDictionary["discMag"] = row["cal_psf_mag"]

            if "transient_object_id" in list(row.keys()):
                thisDictionary[
                    "objectURL"] = "http://star.pst.qub.ac.uk/sne/%(surveyName)s/psdb/candidate/" % locals() + row["transient_object_id"]
            else:
                thisDictionary[
                    "objectURL"] = "http://star.pst.qub.ac.uk/sne/%(surveyName)s/psdb/candidate/" % locals() + row["id"]

            # CLEAN UP IMAGE URLS
            target = row["target"]
            if target:
                id, mjdString, diffId, ippIdet, type = target.split('_')
                thisDictionary["targetImageURL"] = "http://star.pst.qub.ac.uk/sne/%(surveyName)s/site_media/images/data/%(surveyName)s" % locals() + '/' + \
                    str(int(float(mjdString))) + '/' + target + '.jpeg'

            ref = row["ref"]
            if ref:
                id, mjdString, diffId, ippIdet, type = ref.split('_')
                thisDictionary["refImageURL"]  = "http://star.pst.qub.ac.uk/sne/%(surveyName)s/site_media/images/data/%(surveyName)s" % locals() + '/' + \
                    str(int(float(mjdString))) + '/' + ref + '.jpeg'

            diff = row["diff"]
            if diff:
                id, mjdString, diffId, ippIdet, type = diff.split('_')
                thisDictionary["diffImageURL"] = "http://star.pst.qub.ac.uk/sne/%(surveyName)s/site_media/images/data/%(surveyName)s" % locals() + '/' + \
                    str(int(float(mjdString))) + '/' + diff + '.jpeg'

            self.dictList.append(thisDictionary)

        self.log.debug('completed the ``_clean_data_pre_ingest`` method')
        return self.dictList

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*generate colour panstarrs location stamps for regions of transients*

:Author:
    David Young
"""
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import readquery, writequery
from panstamps.downloader import downloader
from panstamps.image import image


class panstarrs_location_stamps(object):
    """
    *The worker class for the panstarrs_location_stamps module*

    **Key Arguments**

    - ``log`` -- logger
    - ``settings`` -- the settings dictionary
    - ``dbConn`` -- dbConn
    - ``transientId`` -- will download for one transient if single ID given. Default *None*


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a panstarrs_location_stamps object, use the following:

    ```python
    from marshallEngine.services import panstarrs_location_stamps
    ps_stamp = panstarrs_location_stamps(
        log=log,
        settings=settings,
        dbConn=dbConn,
        transientId=None
    ).get()
    ```

    """

    def __init__(
            self,
            log,
            dbConn,
            transientId=None,
            settings=False
    ):
        self.log = log
        log.debug("instansiating a new 'panstarrs_location_stamps' object")
        self.settings = settings
        self.dbConn = dbConn
        self.transientId = transientId

        # xt-self-arg-tmpx

        return None

    def get(self):
        """
        *get the panstarrs_location_stamps object*
        """
        self.log.debug('starting the ``get`` method')

        # FOR A SINGLE TRANSIENT
        if self.transientId:
            transientId = self.transientId
            sqlQuery = u"""
                select t.transientBucketId, t.raDeg,t.decDeg from pesstoObjects p, transientBucketSummaries t where p.transientBucketId = t.transientBucketId and t.transientBucketId = %(transientId)s;
            """ % locals()
        # OR THE NEXT 200 TRANSIENTS NEEDING STAMPS
        else:
            # GET NEXT 200 TRANSIENTS NEEDING PANSTARRS STAMPS
            sqlQuery = u"""
                select * from pesstoObjects p, transientBucketSummaries t where (ps1_map is null or ps1_map not in (0,1)) and p.transientBucketId = t.transientBucketId order by t.transientBucketId desc limit 200;
            """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # FOR EACH TRANSIENT DOWNLOAD STAMP TO CACHE DIRECTORY
        downloadDirectoryPath = self.settings[
            "cache-directory"] + "/transients/"

        for row in rows:
            transientBucketId = row["transientBucketId"]
            downloadPath = f"{downloadDirectoryPath}/{transientBucketId}"
            ra = row["raDeg"]
            dec = row["decDeg"]

            fitsPaths, jpegPaths, colorPath = downloader(
                log=self.log,
                settings=self.settings,
                downloadDirectory=downloadPath,
                fits=False,
                jpeg=False,
                arcsecSize=60,
                filterSet='gri',
                color=True,
                singleFilters=False,
                ra=ra,
                dec=dec,
                imageType="stack"  # warp | stack
            ).get()

            # CHECK FOR FAILED IMAGES AND FLAG IN DATABASE
            if len(colorPath) == 0 or not colorPath[0]:
                sqlQuery = u"""
                    update pesstoObjects p, transientBucketSummaries t set p.ps1_map = 0 where p.transientBucketId=t.transientBucketId and (ps1_map is null or ps1_map != 0) and t.decDeg < -40;
                    update pesstoObjects set ps1_map = 2 where transientBucketId = %(transientBucketId)s and ps1_map is null;
                    update pesstoObjects set ps1_map = 2+ps1_map where transientBucketId = %(transientBucketId)s and ps1_map is not null;
                    update pesstoObjects set ps1_map = 0 where transientBucketId = %(transientBucketId)s and ps1_map > 10;
                """ % locals()
                writequery(
                    log=self.log,
                    sqlQuery=sqlQuery,
                    dbConn=self.dbConn
                )
                continue

            source = colorPath[0]
            destination = downloadPath + "/ps1_map_color.jpeg"
            try:
                os.rename(source, destination)

                # DOWNLOAD THE COLOR IMAGE
                myimage = image(
                    log=self.log,
                    settings=self.settings,
                    imagePath=destination,
                    arcsecSize=60,
                    crosshairs=True,
                    transient=False,
                    scale=True,
                    invert=False,
                    greyscale=False
                ).get()

                # UPDATE DATABASE FLAG
                sqlQuery = u"""
                    update pesstoObjects set ps1_map = 1 where transientBucketId = %(transientBucketId)s
                """ % locals()

                writequery(
                    log=self.log,
                    sqlQuery=sqlQuery,
                    dbConn=self.dbConn
                )
            except:
                self.log.warning(
                    "Could not process the image %(destination)s" % locals())

        self.log.debug('completed the ``get`` method')
        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*cache the panstarrs image stamps*

:Author:
    David Young
"""
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import readquery
import requests
from requests.auth import HTTPBasicAuth
import codecs
from fundamentals import fmultiprocess
from fundamentals.mysql import writequery
from ..images import images as baseimages


class images(baseimages):
    """
    *cacher for the panstarrs image stamps*

    **Key Arguments**

    - ``log`` -- logger
    - ``settings`` -- the settings dictionary
    - ``dbConn`` -- the marshall database connection.


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a images object, use the following:

    ```python
    from marshallEngine.feeders.panstarrs import images
    cacher = images(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).cache(limit=1000)  
    ```

    """

    def __init__(
            self,
            log,
            dbConn,
            settings=False
    ):
        self.log = log
        log.debug("instansiating a new 'images' object")
        self.settings = settings
        self.dbConn = dbConn
        self.downloadDirectoryPath = settings[
            "cache-directory"] + "/transients/"

        self.dbSurveyNames = ["pan%%starrs", "ps1"]
        self.stampFlagColumns = {
            "subtracted": None,
            "target": "ps1_target_stamp",
            "reference": None,
            "triplet": None
        }
        self.survey = "ps1"

        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*Baseclass for survey data ingesters*

:Author:
    David Young
"""
from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
import sys
import os
os.environ['TERM'] = 'vt100'
import csv
import requests
from requests.auth import HTTPBasicAuth
from fundamentals.mysql import insert_list_of_dictionaries_into_database_tables, readquery, writequery
from fundamentals import tools


class data(object):
    """
    *This baseclass for the feeder survey data imports*

    **Usage**

    .. todo::

        - create a frankenstein template for importer

    To create a new survey data ingester create a new class using this class as the baseclass:

    ```python
    from ..data import data as basedata
    class data(basedata):
        ....
    ```

    """

    def get_csv_data(
            self,
            url,
            user=False,
            pwd=False):
        """*collect the CSV data from a URL with option to supply basic auth credentials*

        **Key Arguments**

        - ``url`` -- the url to the csv file
        - ``user`` -- basic auth username
        - ``pwd`` -- basic auth password


        **Return**

        - ``csvData`` -- a list of dictionaries from the csv file


        **Usage**

        To get the CSV data for a suvery from a given URL in the marshall settings file run something similar to:

        ```python
        from marshallEngine.feeders.panstarrs.data import data
        ingester = data(
            log=log,
            settings=settings,
            dbConn=dbConn
        )
        csvDicts = ingester.get_csv_data(
            url=settings["panstarrs urls"]["3pi"]["summary csv"],
            user=settings["credentials"]["ps1-3pi"]["username"],
            pwd=settings["credentials"]["ps1-3pi"]["password"]
        )
        ```

        Note you will also be able to access the data via ``ingester.csvDicts`` 

        """
        self.log.debug('starting the ``get_csv_data`` method')

        # DOWNLOAD THE CSV FILE DATA
        try:
            if user:
                response = requests.get(
                    url=url,
                    auth=HTTPBasicAuth(user, pwd)
                )
            else:
                response = requests.get(
                    url=url
                )
            status_code = response.status_code
        except requests.exceptions.RequestException:
            print('HTTP Request failed')
            sys.exit(0)

        if status_code == 502:
            print('HTTP Request failed - status %(status_code)s' % locals())
            print(url)
            self.csvDicts = []
            return self.csvDicts

        if status_code != 200:
            print('HTTP Request failed - status %(status_code)s' % locals())
            sys.exit(0)

        # CONVERT THE RESPONSE TO CSV LIST OF DICTIONARIES
        self.csvDicts = csv.DictReader(
            response.iter_lines(decode_unicode='utf-8'), dialect='excel', delimiter='|', quotechar='"')

        self.log.debug('completed the ``get_csv_data`` method')
        return self.csvDicts

    def _import_to_feeder_survey_table(
            self):
        """*import the list of dictionaries (self.dictList) into the marshall feeder survey table*

        **Return**

        - None


        **Usage**

        ```python
        self._import_to_feeder_survey_table()
        ```

        """
        self.log.debug(
            'starting the ``_import_to_feeder_survey_table`` method')

        if not len(self.dictList):
            return

        # USE dbSettings TO ACTIVATE MULTIPROCESSING
        insert_list_of_dictionaries_into_database_tables(
            dbConn=self.dbConn,
            log=self.log,
            dictList=self.dictList,
            dbTableName=self.fsTableName,
            dateModified=True,
            dateCreated=True,
            batchSize=2500,
            replace=True,
            dbSettings=self.settings["database settings"]
        )

        self.log.debug(
            'completed the ``_import_to_feeder_survey_table`` method')
        return None

    def insert_into_transientBucket(
            self,
            importUnmatched=True,
            updateTransientSummaries=True):
        """*insert objects/detections from the feeder survey table into the transientbucket*

        **Key Arguments**

        - ``importUnmatched`` -- import unmatched (new) transients into the marshall (not wanted in some circumstances)
        - ``updateTransientSummaries`` -- update the transient summaries and lightcurves? Can be True or False, or alternatively a specific transientBucketId


        This method aims to reduce crossmatching and load on the database by:

        1. automatically assign the transientbucket id to feeder survey detections where the object name is found in the transientbukcet (no spatial crossmatch required). Copy matched feeder survey rows to the transientbucket.
        2. crossmatch remaining unique, unmatched sources in feeder survey with sources in the transientbucket. Add associated transientBucketIds to matched feeder survey sources. Copy matched feeder survey rows to the transientbucket.
        3. assign a new transientbucketid to any feeder survey source not matched in steps 1 & 2. Copy these unmatched feeder survey rows to the transientbucket as new transient detections.

        **Return**

        - None


        **Usage**

        ```python
        ingester.insert_into_transientBucket()
        ```

        """
        self.log.debug(
            'starting the ``crossmatch_with_transientBucket`` method')

        fsTableName = self.fsTableName

        # 1. automatically assign the transientbucket id to feeder survey
        # detections where the object name is found in the transientbukcet (no
        # spatial crossmatch required). Copy matched feeder survey rows to the
        # transientbucket.
        self._feeder_survey_transientbucket_name_match_and_import()

        # 2. crossmatch remaining unique, unmatched sources in feeder survey
        # with sources in the transientbucket. Add associated
        # transientBucketIds to matched feeder survey sources. Copy matched
        # feeder survey rows to the transientbucket.
        from HMpTy.mysql import add_htm_ids_to_mysql_database_table
        add_htm_ids_to_mysql_database_table(
            raColName="raDeg",
            declColName="decDeg",
            tableName="transientBucket",
            dbConn=self.dbConn,
            log=self.log,
            primaryIdColumnName="primaryKeyId",
            dbSettings=self.settings["database settings"]
        )
        unmatched = self._feeder_survey_transientbucket_crossmatch()

        # 3. assign a new transientbucketid to any feeder survey source not
        # matched in steps 1 & 2. Copy these unmatched feeder survey rows to
        # the transientbucket as new transient detections.
        if importUnmatched:
            self._import_unmatched_feeder_survey_sources_to_transientbucket(
                unmatched)

        # UPDATE OBSERVATION DATES FROM MJDs
        sqlQuery = "call update_transientbucket_observation_dates()"
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # UPDATE THE TRANSIENT BUCKET SUMMARY TABLE IN THE MARSHALL DATABASE
        if updateTransientSummaries:
            if isinstance(updateTransientSummaries, int) and not isinstance(updateTransientSummaries, bool):
                transientBucketId = updateTransientSummaries
            else:
                transientBucketId = False
            from marshallEngine.housekeeping import update_transient_summaries
            updater = update_transient_summaries(
                log=self.log,
                settings=self.settings,
                dbConn=self.dbConn,
                transientBucketId=transientBucketId
            )
            updater.update()

        self.log.debug(
            'completed the ``crossmatch_with_transientBucket`` method')
        return None

    def _feeder_survey_transientbucket_name_match_and_import(
            self):
        """*automatically assign the transientbucket id to feeder survey detections where the object name is found in the transientbukcet (no spatial crossmatch required). Copy feeder survey rows to the transientbucket.*

        **Return**

        - None


        **Usage**

        ```python
        self._feeder_survey_transientbucket_name_match_and_import()
        ```

        """
        self.log.debug(
            'starting the ``_feeder_survey_transientbucket_name_match_and_import`` method')

        fsTableName = self.fsTableName

        # MATCH TRANSIENT BUCKET IDS WITH NAMES FOUND IN FEEDER TABLE, THEN
        # COPY ROWS TO TRANSIENTBUCKET USING COLUMN MATCH TABLE IN DATABASE
        sqlQuery = """CALL `sync_marshall_feeder_survey_transientBucketId`('%(fsTableName)s');""" % locals(
        )

        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        self.log.debug(
            'completed the ``_feeder_survey_transientbucket_name_match_and_import`` method')
        return None

    def _feeder_survey_transientbucket_crossmatch(
            self):
        """*crossmatch remaining unique, unmatched sources in feeder survey with sources in the transientbucket & copy matched feeder survey rows to the transientbucket*

        **Return**

        - ``unmatched`` -- a list of the unmatched (i.e. new to the marshall) feeder survey surveys

        """
        self.log.debug(
            'starting the ``_feeder_survey_transientbucket_crossmatch`` method')

        fsTableName = self.fsTableName

        # GET THE COLUMN MAP FOR THE FEEDER SURVEY TABLE
        sqlQuery = u"""
            SELECT * FROM marshall_fs_column_map where fs_table_name = '%(fsTableName)s' and transientBucket_column in ('name','raDeg','decDeg','limitingMag')
        """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn,
            quiet=False
        )

        columns = {}
        for row in rows:
            columns[row["transientBucket_column"]] = row["fs_table_column"]

        if "raDeg" not in columns:
            print(f"No coordinates to match in the {fsTableName} table")
            return []

        # BUILD QUERY TO GET UNIQUE UN-MATCHED SOURCES
        fs_name = columns["name"]
        self.fs_name = fs_name
        fs_ra = columns["raDeg"]
        fs_dec = columns["decDeg"]
        if 'limitingMag' in columns:
            fs_lim = columns["limitingMag"]
            limitClause = " and %(fs_lim)s = 0 " % locals()
        else:
            limitClause = ""
        sqlQuery = u"""
            select %(fs_name)s, avg(%(fs_ra)s) as %(fs_ra)s, avg(%(fs_dec)s) as %(fs_dec)s from %(fsTableName)s where ingested = 0 %(limitClause)s and %(fs_ra)s is not null and %(fs_dec)s is not null group by %(fs_name)s 
        """ % locals()

        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn,
            quiet=False
        )

        # STOP IF NO MATCHES
        if not len(rows):
            return []

        # SPLIT INTO BATCHES SO NOT TO OVERWHELM MEMORY
        batchSize = 200
        total = len(rows)
        batches = int(old_div(total, batchSize))
        start = 0
        end = 0
        theseBatches = []
        for i in range(batches + 1):
            end = end + batchSize
            start = i * batchSize
            thisBatch = rows[start:end]
            theseBatches.append(thisBatch)

        unmatched = []
        ticker = 0
        for batch in theseBatches:

            fs_name_list = []
            fs_ra_list = []
            fs_dec_list = []
            fs_name_list = [row[fs_name] for row in batch if row[fs_ra]]
            fs_ra_list = [row[fs_ra] for row in batch if row[fs_ra]]
            fs_dec_list = [row[fs_dec] for row in batch if row[fs_ra]]

            ticker += len(fs_name_list)
            print("Matching %(ticker)s/%(total)s sources in the %(fsTableName)s against the transientBucket table" % locals())

            # CONESEARCH TRANSIENT BUCKET FOR PRE-KNOWN SOURCES FROM OTHER
            # SURVEYS
            from HMpTy.mysql import conesearch
            cs = conesearch(
                log=self.log,
                dbConn=self.dbConn,
                tableName="transientBucket",
                columns="transientBucketId, name",
                ra=fs_ra_list,
                dec=fs_dec_list,
                radiusArcsec=3.5,
                separations=True,
                distinct=True,
                sqlWhere="masterIDFlag=1",
                closest=True
            )
            matchIndies, matches = cs.search()

            # CREATE SQL QUERY TO UPDATE MATCHES IN FS TABLE WITH MATCHED
            # TRANSIENTBUCKET IDs
            updates = []
            originalList = matches.list
            originalTotal = len(originalList)

            print("Adding %(originalTotal)s new %(fsTableName)s transient detections to the transientBucket table" % locals())
            if originalTotal:
                updates = []
                updates[:] = ["update " + fsTableName + " set transientBucketId = " + str(o['transientBucketId']) +
                              " where " + fs_name + " = '" + str(fs_name_list[m]) + "' and transientBucketId is null;" for m, o in zip(matchIndies, originalList)]
                updates = ("\n").join(updates)
                writequery(
                    log=self.log,
                    sqlQuery=updates,
                    dbConn=self.dbConn
                )

            # RETURN UNMATCHED TRANSIENTS
            for i, v in enumerate(fs_name_list):
                if i not in matchIndies:
                    unmatched.append(v)

        # COPY MATCHED ROWS TO TRANSIENTBUCKET
        self._feeder_survey_transientbucket_name_match_and_import()

        self.log.debug(
            'completed the ``_feeder_survey_transientbucket_crossmatch`` method')
        return unmatched

    def _import_unmatched_feeder_survey_sources_to_transientbucket(
            self,
            unmatched):
        """*assign a new transientbucketid to any feeder survey source not yet matched in steps. Copy these unmatched feeder survey rows to the transientbucket as new transient detections.*

        **Key Arguments**

        - ``unmatched`` -- the remaining unmatched feeder survey object names.

        """
        self.log.debug(
            'starting the ``_import_unmatched_feeder_survey_sources_to_transientbucket`` method')

        if not len(unmatched):
            return None

        fsTableName = self.fsTableName
        fs_name = self.fs_name

        # READ MAX TRANSIENTBUCKET ID FROM TRANSIENTBUCKET
        sqlQuery = u"""
            select max(transientBucketId) as maxId from transientBucket
        """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        if not len(rows) or not rows[0]["maxId"]:
            maxId = 1
        else:
            maxId = rows[0]["maxId"] + 1

        # ADD NEW TRANSIENTBUCKETIDS TO FEEDER SURVEY TABLE
        updates = []
        newTransientBucketIds = []
        for u in unmatched:
            update = "update " + fsTableName + " set transientBucketId = " + \
                str(maxId) + " where " + fs_name + " = '" + str(u) + "';"
            updates.append(update)
            newTransientBucketIds.append(str(maxId))
            maxId += 1
        updates = ("\n").join(updates)
        writequery(
            log=self.log,
            sqlQuery=updates,
            dbConn=self.dbConn
        )

        # COPY FEEDER SURVEY ROWS TO TRANSIENTBUCKET
        self._feeder_survey_transientbucket_name_match_and_import()

        # SET THE MASTER ID FLAG FOR ALL NEW TRANSIENTS IN THE TRANSIENTBUCKET
        newTransientBucketIds = (",").join(newTransientBucketIds)
        sqlQuery = """UPDATE transientBucket t
                            JOIN
                        (SELECT 
                            transientBucketId, MIN(primaryKeyId) AS minpk
                        FROM
                            transientBucket
                        WHERE
                            transientBucketId IN (%(newTransientBucketIds)s)
                        GROUP BY transientBucketId) tmin ON t.primaryKeyId = tmin.minpk 
                    SET 
                        masterIDFlag = 1;""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        self.log.debug(
            'completed the ``_import_unmatched_feeder_survey_sources_to_transientbucket`` method')
        return None

    # use the tab-trigger below for new method
    def clean_up(
            self):
        """*A few tasks to finish off the ingest*

        **Key Arguments:**
            # -

        **Return:**
            - None

        **Usage:**

        ```python
        usage code 
        ```

        ---

        ```eval_rst
        .. todo::

            - add usage info
            - create a sublime snippet for usage
            - write a command-line tool for this method
            - update package tutorial with command-line tool info if needed
        ```
        """
        self.log.debug('starting the ``clean_up`` method')

        sqlQueries = [
            "insert into sherlock_classifications (transient_object_id) select distinct transientBucketId from transientBucketSummaries ON DUPLICATE KEY UPDATE  transient_object_id = transientBucketId;",
            "CALL update_transient_akas(1); "
        ]

        for sqlQuery in sqlQueries:
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn
            )

        self.log.debug('completed the ``clean_up`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/bin/env python
# encoding: utf-8
"""
Documentation for marshallEngine can be found here: http://marshallEngine.readthedocs.org

Usage:
    marshall init
    marshall clean [-s <pathToSettingsFile>]
    marshall import <survey> [<withInLastDay>] [-s <pathToSettingsFile>]
    marshall lightcurve <transientBucketId> [-s <pathToSettingsFile>]
    marshall refresh <transientBucketId>  [-s <pathToSettingsFile>]

Options:
    init                  setup the marshallEngine settings file for the first time
    clean                 preform cleanup tasks like updating transient summaries table
    import                import data, images, lightcurves from a feeder survey
    refresh               update the cached metadata for a given transient
    lightcurve            generate a lightcurve for a transient in the marshall database
    transientBucketId     the transient ID from the database
    survey                name of survey to import [panstarrs|atlas|useradded]
    withInLastDay         import transient detections from the last N days (Default 30)

    -h, --help                              show this help message
    -v, --version                           show version
    -s, --settings <pathToSettingsFile>     the settings file
"""
from __future__ import print_function
import sys
import os
os.environ['TERM'] = 'vt100'
import readline
import glob
import pickle
from docopt import docopt
from fundamentals import tools, times
from subprocess import Popen, PIPE, STDOUT


def tab_complete(text, state):
    return (glob.glob(text + '*') + [None])[state]


def main(arguments=None):
    """
    *The main function used when `cl_utils.py` is run as a single script from the cl, or when installed as a cl command*
    """
    # setup the command-line util settings
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="WARNING",
        options_first=False,
        projectName="marshall",
        defaultSettingsFile=True
    )
    arguments, settings, log, dbConn = su.setup()

    # tab completion for raw_input
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind("tab: complete")
    readline.set_completer(tab_complete)

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    a = {}
    for arg, val in list(arguments.items()):
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        a[varname] = val
        if arg == "--dbConn":
            dbConn = val
            a["dbConn"] = val
        log.debug('%s = %s' % (varname, val,))

    ## START LOGGING ##
    startTime = times.get_now_sql_datetime()
    log.info(
        '--- STARTING TO RUN THE cl_utils.py AT %s' %
        (startTime,))

    init = a["init"]
    clean = a["clean"]
    iimport = a["import"]
    lightcurve = a["lightcurve"]
    transientBucketId = a["transientBucketId"]
    survey = a["survey"]
    withInLastDay = a["withInLastDay"]
    settingsFlag = a["settingsFlag"]

    # set options interactively if user requests
    if "interactiveFlag" in a and a["interactiveFlag"]:

        # load previous settings
        moduleDirectory = os.path.dirname(__file__) + "/resources"
        pathToPickleFile = "%(moduleDirectory)s/previousSettings.p" % locals()
        try:
            with open(pathToPickleFile):
                pass
            previousSettingsExist = True
        except:
            previousSettingsExist = False
        previousSettings = {}
        if previousSettingsExist:
            previousSettings = pickle.load(open(pathToPickleFile, "rb"))

        # x-raw-input
        # x-boolean-raw-input
        # x-raw-input-with-default-value-from-previous-settings

        # save the most recently used requests
        pickleMeObjects = []
        pickleMe = {}
        theseLocals = locals()
        for k in pickleMeObjects:
            pickleMe[k] = theseLocals[k]
        pickle.dump(pickleMe, open(pathToPickleFile, "wb"))

    if a["init"]:
        from os.path import expanduser
        home = expanduser("~")
        filepath = home + "/.config/marshallEngine/marshallEngine.yaml"
        try:
            cmd = """open %(filepath)s""" % locals()
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        except:
            pass
        try:
            cmd = """start %(filepath)s""" % locals()
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        except:
            pass
        return

    # CALL FUNCTIONS/OBJECTS
    # DEFAULT VALUES
    if not withInLastDay:
        withInLastDay = 30

    # CALL FUNCTIONS/OBJECTS
    if clean:
        # RESCUE ORPHANED TRANSIENTS - NO MASTER ID FLAG
        print("running stored procedures ...")
        from fundamentals.mysql import writequery

        procedureNames = [
            "update_transients_with_no_masteridflag()",
            "insert_new_transients_into_transientbucketsummaries()",
            "resurrect_objects()",
            "update_sherlock_xmatch_counts()",
            "update_inbox_auto_archiver()",
            "update_transient_akas(0)"
        ]

        # CALL EACH PROCEDURE
        for p in procedureNames:
            sqlQuery = "CALL %(p)s;" % locals()
            print(f'Calling the `{p}` MySQL procedure')
            writequery(
                log=log,
                sqlQuery=sqlQuery,
                dbConn=dbConn,
            )

        # UPDATE THE TRANSIENT BUCKET SUMMARY TABLE IN THE MARSHALL DATABASE
        from marshallEngine.housekeeping import update_transient_summaries
        updater = update_transient_summaries(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).update()

    if iimport:
        if survey.lower() == "panstarrs":
            from marshallEngine.feeders.panstarrs.data import data
            from marshallEngine.feeders.panstarrs import images
        if survey.lower() == "atlas":
            from marshallEngine.feeders.atlas.data import data
            from marshallEngine.feeders.atlas import images
        if survey.lower() == "useradded":
            from marshallEngine.feeders.useradded.data import data
            from marshallEngine.feeders.useradded import images
        if survey.lower() == "tns":
            from marshallEngine.feeders.tns.data import data
            from marshallEngine.feeders.tns import images
        if survey.lower() == "ztf":
            from marshallEngine.feeders.ztf.data import data
            from marshallEngine.feeders.ztf import images
        if survey.lower() == "atels" or survey.lower() == "atel":
            from marshallEngine.feeders.atels.data import data
            from marshallEngine.feeders.atels import images
        ingester = data(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).ingest(withinLastDays=withInLastDay)
        cacher = images(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).cache(limit=3000)

        from marshallEngine.services import panstarrs_location_stamps
        ps_stamp = panstarrs_location_stamps(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).get()

    if lightcurve:
        from marshallEngine.lightcurves import marshall_lightcurves
        lc = marshall_lightcurves(
            log=log,
            dbConn=dbConn,
            settings=settings,
            transientBucketIds=transientBucketId
        )
        filepath = lc.plot()
        print("The lightcurve plot for transient %(transientBucketId)s can be found here: %(filepath)s" % locals())

    if a["refresh"]:
        from marshallEngine.housekeeping import update_transient_summaries
        updater = update_transient_summaries(
            log=log,
            settings=settings,
            dbConn=dbConn,
            transientBucketId=transientBucketId
        ).update()

    if "dbConn" in locals() and dbConn:
        dbConn.commit()
        dbConn.close()
    ## FINISH LOGGING ##
    endTime = times.get_now_sql_datetime()
    runningTime = times.calculate_time_difference(startTime, endTime)
    log.info('-- FINISHED ATTEMPT TO RUN THE cl_utils.py AT %s (RUNTIME: %s) --' %
             (endTime, runningTime, ))

    return

if __name__ == '__main__':
    main()



#!/usr/local/bin/python
# encoding: utf-8
"""
*import the tns stream into the marshall*

:Author:
    David Young
"""
from builtins import zip
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
from transientNamer import search
from fundamentals.mysql import insert_list_of_dictionaries_into_database_tables


class data(basedata):
    """
    *Import the tns transient data into the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``dbConn`` -- the marshall database connection
    - ``settings`` -- the settings dictionary


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a data object, use the following:

    ```python
    from marshallEngine.feeders.tns.data import data
    ingester = data(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).ingest()   
    ```

    """

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn
        self.fsTableName = "tns_sources"
        self.survey = "tns"

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays=False):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments**

        - ``withinLastDays`` -- note this will be handle by the transientNamer import to the database

        """
        self.log.debug('starting the ``ingest`` method')

        # UPDATE THE TNS SPECTRA TABLE WITH EXTRA INFOS
        from fundamentals.mysql import writequery
        sqlQuery = """CALL `update_tns_tables`();""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # PARSE TNS
        tns = search(
            log=self.log,
            discInLastDays=withinLastDays
        )

        lists = [tns.sources, tns.photometry, tns.files, tns.spectra]
        tableNames = ["tns_sources", "tns_photometry",
                      "tns_files", "tns_spectra"]

        for l, t in zip(lists, tableNames):
            # USE dbSettings TO ACTIVATE MULTIPROCESSING - INSERT LIST OF
            # DICTIONARIES INTO DATABASE
            insert_list_of_dictionaries_into_database_tables(
                dbConn=self.dbConn,
                log=self.log,
                dictList=l,
                dbTableName=t,
                dateModified=True,
                dateCreated=True,
                batchSize=2500,
                replace=True,
                dbSettings=self.settings["database settings"]
            )

        # INSERT THE SOURCES TABLE
        self.insert_into_transientBucket()

        # NOW THE SPECTRA TABLE
        self.fsTableName = "tns_spectra"
        self.survey = "tns"
        self.insert_into_transientBucket(importUnmatched=False)

        # NOW THE PHOTOMETRY TABLE
        self.fsTableName = "tns_photometry"
        self.survey = "tns"
        self.insert_into_transientBucket(importUnmatched=False)

        # ALSO MATCH NEW ASTRONOTES
        sqlQuery = """CALL sync_marshall_feeder_survey_transientBucketId('astronotes_transients');""" % locals(
        )
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.debug('completed the ``ingest`` method')
        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*import the ZTF stream into the marshall*

:Author:
    David Young
"""
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
from fundamentals.mysql import writequery


class data(basedata):
    """
    *Import the ZTF transient data into the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``dbConn`` -- the marshall database connection
    - ``settings`` -- the settings dictionary


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    To initiate a data object, use the following:

    ```python
    from marshallEngine.feeders.ztf.data import data
    ingester = data(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).ingest(withinLastDays=withInLastDay)   
    ```

    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn

        self.fsTableName = "fs_ztf"
        self.survey = "ZTF"

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments**

        - ``withinLastDays`` -- within the last number of days. *Default: 50*

        """
        self.log.debug('starting the ``ingest`` method')

        allLists = []

        sqlQuery = """call update_fs_ztf()""" % locals()
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # self._import_to_feeder_survey_table()
        self.insert_into_transientBucket()

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.debug('completed the ``ingest`` method')
        return None



#!/usr/bin/env python
# encoding: utf-8
"""
*import the atels stream into the marshall*

:Author:
    David Young

:Date Created:
    January 12, 2021
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
import git
from fundamentals.mysql import writequery


class data(basedata):
    """
    *Import the atels transient data into the marshall database*

    **Key Arguments:**
        - ``log`` -- logger
        - ``dbConn`` -- the marshall database connection
        - ``settings`` -- the settings dictionary

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a data object, use the following:

        .. code-block:: python 

            from marshallEngine.feeders.atels.data import data
            ingester = data(
                log=log,
                settings=settings,
                dbConn=dbConn
            ).ingest(withinLastDays=withInLastDay)   
    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn

        self.fsTableName = "atel_coordinates"
        self.survey = "atel"
        self.dictList = []

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments:**
            - ``withinLastDays`` -- within the last number of days. *Default: 50*
        """
        self.log.debug('starting the ``ingest`` method')

        allLists = []

        self.download_new_atels()
        self.parse_atels_to_database()

        self.insert_into_transientBucket(
            importUnmatched=False,
            updateTransientSummaries=True)

        self.fsTableName = "atel_names"
        self.insert_into_transientBucket(
            importUnmatched=False,
            updateTransientSummaries=True)

        sqlQuery = f"""CALL insert_atel_titles_to_comments();"""
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.debug('completed the ``ingest`` method')
        return None

    def update_git_repo(
            self):
        """*update the atel data git repo (if it exists)*
        ```
        """
        self.log.debug('starting the ``update_git_repo`` method')

        # TEST IF atel-directory IS A GIT REPO
        atelDir = self.settings["atel-directory"]

        # FIRST TRY RAW PATH
        try:
            repo = git.Repo(atelDir)
        except:
            repo = False

        # IF THIS FAILS TRY PARENT DIRECTORY
        if repo == False:
            try:
                repo = os.path.dirname(atelDir)
                repo = git.Repo(repo)
            except:
                repo = False

        # PUSH/PULL CHAGES
        if repo != False:
            repo.git.add(update=True)
            repo.index.commit("adding atels")
            o = repo.remotes.origin
            o.pull()
            o.push()
            o.pull()

        self.log.debug('completed the ``update_git_repo`` method')
        return None

    def download_new_atels(
            self):
        """*download new atel html files*
        """
        self.log.debug('starting the ``download_new_atels`` method')

        self.update_git_repo()
        from atelParser import download
        atels = download(
            log=self.log,
            settings=self.settings
        )
        atelsToDownload = atels.get_list_of_atels_still_to_download()
        if "test" in self.settings and self.settings["test"] == True:
            atelsToDownload = [atelsToDownload[-1]]
            atels.maxsleep = 1
        else:
            atels.maxsleep = 20
        atels.download_list_of_atels(atelsToDownload)
        self.update_git_repo()

        self.log.debug('completed the ``download_new_atels`` method')
        return None

    def parse_atels_to_database(
            self):
        """*parse content of atels to the marshall database.*

        This populates the `atel_coordinates`, `atel_fullcontent` and `atel_names` database tables.
        """
        self.log.debug('starting the ``parse_atels_to_database`` method')

        from atelParser import mysql
        parser = mysql(
            log=self.log,
            settings=self.settings
        )
        parser.atels_to_database()
        parser.parse_atels()
        parser.populate_htm_columns()

        self.log.debug('completed the ``parse_atels_to_database`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*import the useradded stream into the marshall*

:Author:
    David Young
"""
from __future__ import print_function
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ..data import data as basedata
from astrocalc.times import now
from fundamentals.mysql import readquery
from marshallEngine.housekeeping import update_transient_summaries


class data(basedata):
    """
    *Import the useradded transient data into the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``dbConn`` -- the marshall database connection
    - ``settings`` -- the settings dictionary


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To initiate a data object, use the following:

    ```python
    from marshallEngine.feeders.useradded.data import data
    ingester = data(
        log=log,
        settings=settings,
        dbConn=dbConn
    ).ingest(withinLastDays=withInLastDay)
    ```

    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
    ):
        self.log = log
        log.debug("instansiating a new 'data' object")
        self.settings = settings
        self.dbConn = dbConn

        self.fsTableName = "fs_user_added"
        self.survey = "useradded"

        # xt-self-arg-tmpx

        return None

    def ingest(
            self,
            withinLastDays):
        """*Ingest the data into the marshall feeder survey table*

        **Key Arguments**

        - ``withinLastDays`` -- within the last number of days. *Default: 50*

        """
        self.log.info('starting the ``ingest`` method')

        allLists = []

        self.dictList = allLists
        self._import_to_feeder_survey_table()

        self.insert_into_transientBucket(
            updateTransientSummaries=False)

        sqlQuery = u"""
            select transientBucketId from fs_user_added where transientBucketId is not null order by dateCreated desc limit 1
        """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        if len(rows):
            transientBucketId = rows[0]["transientBucketId"]
            print(transientBucketId)
        else:
            transientBucketId = False

        # UPDATE THE TRANSIENT BUCKET SUMMARY TABLE IN THE MARSHALL DATABASE
        updater = update_transient_summaries(
            log=self.log,
            settings=self.settings,
            dbConn=self.dbConn,
            transientBucketId=transientBucketId
        )
        updater.update()

        # CLEAN UP TASKS TO MAKE THE TICKET UPDATE
        self.clean_up()

        self.log.info('completed the ``ingest`` method')
        return None

    def _clean_data_pre_ingest(
            self,
            surveyName,
            withinLastDays=False):
        """*clean up the list of dictionaries containing the useradded data, pre-ingest*

        **Key Arguments**

        - ``surveyName`` -- the useradded survey name
        -  ``withinLastDays`` -- the lower limit of observations to include (within the last N days from now). Default *False*, i.e. no limit


        **Return**

        - ``dictList`` -- the cleaned list of dictionaries ready for ingest


        **Usage**

        To clean the data from the useradded survey:

        ```python
        dictList = ingesters._clean_data_pre_ingest(surveyName="useradded")
        ```

        Note you will also be able to access the data via ``ingester.dictList``

        """
        self.log.info('starting the ``_clean_data_pre_ingest`` method')

        self.dictList = []

        # CALC MJD LIMIT
        if withinLastDays:
            mjdLimit = now(
                log=self.log
            ).get_mjd() - float(withinLastDays)

        for row in self.csvDicts:
            # IF NOW IN THE LAST N DAYS - SKIP
            if withinLastDays and float(row["mjd_obs"]) < mjdLimit:
                continue

            # MASSAGE THE DATA IN THE INPT FORMAT TO WHAT IS NEEDED IN THE
            # FEEDER SURVEY TABLE IN THE DATABASE
            thisDictionary = {}
            # thisDictionary["candidateID"] = row["ps1_designation"]
            # ...

            self.dictList.append(thisDictionary)

        self.log.info('completed the ``_clean_data_pre_ingest`` method')
        return self.dictList

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*Update the transient summaries table in the marshall database with the top-level metadata*

:Author:
    David Young
"""
from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import writequery, readquery
from astropy import units as u
from astropy.coordinates import SkyCoord
from fundamentals.mysql import insert_list_of_dictionaries_into_database_tables
from astrocalc.distances import converter
import numpy as np
from HMpTy.mysql import add_htm_ids_to_mysql_database_table


class update_transient_summaries(object):
    """
    *Update the transient summaries table in the marshall database*

    **Key Arguments**

    - ``log`` -- logger
    - ``settings`` -- the settings dictionary
    - ``dbConn`` -- the marshall database connection
    - ``transientBucketId`` -- a single transientBucketId to update transientBucketId. Default *False* (i.e. update all)


    **Usage**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    To initiate a update_transient_summaries object, use the following:

    .. todo::

        - add a tutorial about ``update_transient_summaries`` to documentation

    ```python
    from marshallEngine.housekeeping import update_transient_summaries
    updater = update_transient_summaries(
        log=log,
        settings=settings,
        dbConn=dbConn,
        transientBucketId=False
    ).update()
    ```

    """
    # Initialisation

    def __init__(
            self,
            log,
            dbConn,
            settings=False,
            transientBucketId=False
    ):
        self.log = log
        log.debug("instansiating a new 'update_transient_summaries' object")
        self.settings = settings
        self.transientBucketIds = []
        self.transientBucketId = transientBucketId

        # xt-self-arg-tmpx

        self.dbConn = dbConn

        if self.transientBucketId:
            print(
                "updating transient summaries table for %(transientBucketId)s" % locals())
            # UPDATE TRANSIENT BUCKET SUMMARIES (IN MYSQL)
            sqlQuery = "call update_single_transientbucket_summary(%(transientBucketId)s)" % locals(
            )
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn
            )
        else:
            # UPDATE OBSERVATION DATES FROM MJDs
            sqlQuery = "call update_transientbucket_observation_dates()"
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn
            )

            print("updating transient summaries table")
            # UPDATE TRANSIENT BUCKET SUMMARIES (IN MYSQL)
            sqlQuery = "call update_transientbucketsummaries()"
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn
            )

            print("updating sherlock crossmatches table")
            sqlQuery = "call update_sherlock_crossmatches()"
            writequery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn
            )

        return None

    def update(self):
        """
        *Update the transient summaries table in the marshall database*

        **Return**

        - ``update_transient_summaries``


        **Usage**



        ```python
        from marshallEngine.housekeeping import update_transient_summaries
        updater = update_transient_summaries(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).update()
        ```
        """
        self.log.debug('starting the ``get`` method')

        self._add_galactic_coords()
        self._add_distances()

        # UPDATE LIGHTCURVE PLOTS

        from marshallEngine.lightcurves import marshall_lightcurves
        lc = marshall_lightcurves(
            log=self.log,
            dbConn=self.dbConn,
            settings=self.settings,
            transientBucketIds=self.transientBucketIds
        )
        lc.plot()

        transientBucketIds = ('","').join(str(x)
                                          for x in self.transientBucketIds)

        # RESET UPDATE-NEEDED FLAG
        sqlQuery = """update transientBucketSummaries set updateNeeded = 0 where updateNeeded = 2 and transientBucketId in ("%(transientBucketIds)s")""" % locals(
        )
        writequery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn
        )

        # ADD HTM IDs
        self._update_htm_columns()

        self.log.debug('completed the ``get`` method')
        return None

    def _add_galactic_coords(
            self):
        """*add galactic coordinates to the summary table*

        **Usage**

        ```python
        from marshallEngine.housekeeping import update_transient_summaries
        updater = update_transient_summaries(
            log=log,
            settings=settings,
            dbConn=dbConn
        )
        updater._add_galactic_coords()
        ```

        """
        self.log.debug('starting the ``_add_galactic_coords`` method')

        # SELECT THE TRANSIENTS NEEDING UPDATED
        extra = ""
        if self.transientBucketId:
            thisId = self.transientBucketId
            extra = "and transientBucketId = %(thisId)s" % locals()

        sqlQuery = u"""
            select raDeg, decDeg, transientBucketId from transientBucketSummaries where updateNeeded = 2 %(extra)s order by transientBucketId desc
        """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn,
            quiet=False
        )

        total = len(rows)
        # STOP IF NOTHING TO UPDATE
        if not total:
            return

        if total > 1000:
            print(
                """%(total)s transients need updated - updating the next 1000""" % locals())
            rows = np.random.choice(rows, size=1000, replace=False, p=None)

        # CREATE 3 LISTS - RA, DEC, ID
        raDeg = []
        raDeg[:] = [r["raDeg"] * u.degree for r in rows]
        decDeg = []
        decDeg[:] = [r["decDeg"] * u.degree for r in rows]
        ids = []
        ids[:] = [r["transientBucketId"] for r in rows]
        self.transientBucketIds = ids

        # DETERMINE GALACTIC COORDINATES
        c = SkyCoord(ra=raDeg, dec=decDeg, frame='icrs')
        l = c.galactic.l.deg
        b = c.galactic.b.deg

        # CREATE A LIST OF DICTIONARIES
        dictList = []
        for gl, gb, tid in zip(l, b, ids):
            dictList.append({
                "glat": gb,
                "glon": gl,
                "transientBucketId": tid
            })

        if not len(dictList):
            return

        # USE dbSettings TO ACTIVATE MULTIPROCESSING - INSERT LIST OF
        # DICTIONARIES INTO DATABASE
        insert_list_of_dictionaries_into_database_tables(
            dbConn=self.dbConn,
            log=self.log,
            dictList=dictList,
            dbTableName="transientBucketSummaries",
            dateModified=False,
            dateCreated=False,
            batchSize=2500,
            replace=True,
            dbSettings=self.settings["database settings"]
        )

        self.log.debug('completed the ``_add_galactic_coords`` method')
        return None

    def _add_distances(
            self):
        """*Add a distance measurement from the best redshift if the transient does not already have a distance measurement*

        **Usage**

        ```python
        from marshallEngine.housekeeping import update_transient_summaries
        updater = update_transient_summaries(
            log=log,
            settings=settings,
            dbConn=dbConn
        )
        updater._add_distances()
        ```

        """
        self.log.debug('starting the ``_add_distances`` method')

        extra = ""
        if self.transientBucketId:
            thisId = self.transientBucketId
            extra = "and transientBucketId = %(thisId)s" % locals()

        # SELECT THE TRANSIENTS NEEDING UPDATED
        sqlQuery = u"""
            select best_redshift, transientBucketId from transientBucketSummaries where best_redshift is not null and distanceMpc is null and best_redshift > 0.001 %(extra)s
        """ % locals()
        rows = readquery(
            log=self.log,
            sqlQuery=sqlQuery,
            dbConn=self.dbConn,
            quiet=False
        )

        # CONVERT REDSHIFT TO DISTANCE
        c = converter(log=self.log)
        dictList = []
        for r in rows:
            dists = c.redshift_to_distance(
                z=r["best_redshift"],
                WM=0.3,
                WV=0.7,
                H0=70.0
            )
            dmod = dists["dmod"]
            dl_mpc = dists["dl_mpc"]
            da_scale = dists["da_scale"]
            da_mpc = dists["da_mpc"]
            dcmr_mpc = dists["dcmr_mpc"]
            dictList.append({
                "transientBucketId": r["transientBucketId"],
                "distanceMpc": dl_mpc
            })

        if not len(dictList):
            return

        # USE dbSettings TO ACTIVATE MULTIPROCESSING - INSERT LIST OF
        # DICTIONARIES INTO DATABASE
        insert_list_of_dictionaries_into_database_tables(
            dbConn=self.dbConn,
            log=self.log,
            dictList=dictList,
            dbTableName="transientBucketSummaries",
            dateModified=False,
            dateCreated=False,
            batchSize=2500,
            replace=True,
            dbSettings=self.settings["database settings"]
        )

        self.log.debug('completed the ``_add_distances`` method')
        return None

    def _update_htm_columns(
            self):
        """*update the htm columns in the transientSummaries table so we can crossmatch if needed*
        """
        self.log.debug('starting the ``_update_htm_columns`` method')

        add_htm_ids_to_mysql_database_table(
            raColName="raDeg",
            declColName="decDeg",
            tableName="transientBucketSummaries",
            dbConn=self.dbConn,
            log=self.log,
            primaryIdColumnName="transientBucketId",
            dbSettings=self.settings["database settings"]
        )

        self.log.debug('completed the ``_update_htm_columns`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*cache the panstarrs image stamps*

:Author:
    David Young
"""
from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import zip
from builtins import object
from past.utils import old_div
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from fundamentals.mysql import readquery
import requests
from requests.auth import HTTPBasicAuth
import codecs
from fundamentals import fmultiprocess
from fundamentals.mysql import writequery


class images(object):
    """
    *The base class for the feeder image cachers*

    **Usage**

    To create a new survey image cacher create a new class using this class as the baseclass:

    ```python
    from ..images import images as baseimages
    class images(baseimages):
        ....
    ```

    """

    def cache(
            self,
            limit=1000):
        """*cache the image for the requested survey*

        **Key Arguments**

        - ``limit`` -- limit the number of transients in the list so not to piss-off survey owners by downloading everything in one go.


        **Usage**

        ```python
        from marshallEngine.feeders.panstarrs import images
        cacher = images(
            log=log,
            settings=settings,
            dbConn=dbConn
        ).cache(limit=1000)
        ```

        """
        self.log.debug('starting the ``cache`` method')

        # THESE SURVEY DON'T HAVE IMAGES - PASS
        if self.survey in ["tns", "atel", "atels"]:
            return

        transientBucketIds, subtractedUrls, targetUrls, referenceUrls, tripletUrls = self._list_images_needing_cached()
        leng = len(transientBucketIds)
        survey = self.survey

        if not leng:
            print("All _new_ images are cached for the %(survey)s survey" % locals())

        else:

            if leng > limit:
                print(
                    "Downloading image stamps for the next %(limit)s transients of %(leng)s remaining for %(survey)s" % locals())
            else:
                print(
                    "Downloading image stamps for the remaining %(leng)s transients for %(survey)s" % locals())
            subtractedStatus, targetStatus, referenceStatus, tripletStatus = self._download(
                transientBucketIds=transientBucketIds[:limit],
                subtractedUrls=subtractedUrls[:limit],
                targetUrls=targetUrls[:limit],
                referenceUrls=referenceUrls[:limit],
                tripletUrls=tripletUrls[:limit]
            )
            self._update_database()

        transientBucketIds, subtractedUrls, targetUrls, referenceUrls, tripletUrls = self._list_images_needing_cached(
            failedImage=True)
        leng = len(transientBucketIds)

        if not leng:
            print("All images are cached for the %(survey)s survey" % locals())

        else:
            print("Downloading image stamps for the next %(limit)s transients of %(leng)s remaining for %(survey)s - previously failed" % locals())
            subtractedStatus, targetStatus, referenceStatus, tripletStatus = self._download(
                transientBucketIds=transientBucketIds[:limit],
                subtractedUrls=subtractedUrls[:limit],
                targetUrls=targetUrls[:limit],
                referenceUrls=referenceUrls[:limit],
                tripletUrls=tripletUrls[:limit]
            )
            self._update_database()

        self.log.debug('completed the ``cache`` method')
        return None

    def _list_images_needing_cached(
            self,
            failedImage=False):
        """*get lists of the transientBucketIds and images needing cached for those transients*

        **Key Arguments**

        - ``failedImage`` -- second pass attempt to download alternative image for transients


        **Return**

        - ``transientBucketIds, subtractedUrls, targetUrls, referenceUrls, tripletUrls`` -- synced lists of transientBucketIds, subtracted-, target-, reference- and triplet-image urls. All lists are the same size.

        """
        self.log.debug('starting the ``_list_images_needing_cached`` method')

        subtractedUrls, targetUrls, referenceUrls, tripletUrls = [], [], [], []
        for imageType, v in list(self.stampFlagColumns.items()):
            if not v:
                continue
            imageUrl = imageType + "ImageUrl"
            # CREATE THE STAMP WHERE CLAUSE
            if not failedImage:
                stampWhere = v + " IS NULL "
            else:
                stampWhere = v + " = 2 "

            # CREATE THE SURVEY WHERE CLAUSE
            dbSurveyNames = "t.survey LIKE '%%" + \
                ("%%' OR t.survey LIKE '%%").join(self.dbSurveyNames) + "%%'"

            # NOW GENERATE SQL TO GET THE URLS OF STAMPS NEEDING DOWNLOADED
            if self.survey == "useradded":
                sqlQuery = u"""
                SELECT 
        a.transientBucketId, a.%(imageUrl)s 
    FROM
        transientBucket a
            JOIN
        (SELECT 
            MIN(magnitude) AS mag, transientBucketId
        FROM
            transientBucket
        WHERE
            magnitude IS NOT NULL
                AND %(imageUrl)s IS NOT NULL
                AND transientBucketId in (select transientBucketId from fs_user_added)
                AND transientBucketId IN (SELECT 
                    transientBucketId
                FROM
                    pesstoObjects
                WHERE
                    %(stampWhere)s and limitingMag = 0)
        GROUP BY transientBucketId
        ORDER BY transientBucketId) AS b ON a.transientBucketId = b.transientBucketId
            AND a.magnitude = b.mag
            WHERE limitingMag = 0
        GROUP BY transientBucketId;
            """ % locals()
            else:
                sqlQuery = u"""
                    SELECT 
    MIN(magnitude) AS mag, t.transientBucketId, %(imageUrl)s
FROM
    transientBucket t,
    pesstoObjects p
WHERE
    t.magnitude IS NOT NULL
        AND t.%(imageUrl)s IS NOT NULL
        AND p.%(stampWhere)s
        AND t.transientbucketId = p.transientbucketId
        AND (%(dbSurveyNames)s)
        AND t.limitingMag = 0 group by t.transientBucketId;""" % locals()

            if failedImage:
                sqlQuery = sqlQuery.replace("AND a.magnitude = b.mag", "").replace(
                    "GROUP BY a.transientBucketId;", "")

            rows = readquery(
                log=self.log,
                sqlQuery=sqlQuery,
                dbConn=self.dbConn,
            )

            # SPLIT URLS INTO STAMP TYPES AND ORDER ALONGSIDE
            # TRANSIENTBUKCETIDs
            transientBucketIds = []
            for row in rows:
                transientBucketIds.append(row["transientBucketId"])
                if imageType == "subtracted":
                    subtractedUrls.append(row["subtractedImageUrl"])
                if imageType == "target":
                    targetUrls.append(row["targetImageUrl"])
                if imageType == "reference":
                    referenceUrls.append(row["referenceImageUrl"])
                if imageType == "triplet":
                    tripletUrls.append(row["tripletImageUrl"])

        for imageType, v in list(self.stampFlagColumns.items()):
            if not v:
                if imageType == "subtracted":
                    subtractedUrls = [None] * len(transientBucketIds)
                if imageType == "target":
                    targetUrls = [None] * len(transientBucketIds)
                if imageType == "reference":
                    referenceUrls = [None] * len(transientBucketIds)
                if imageType == "triplet":
                    tripletUrls = [None] * len(transientBucketIds)

        self.log.debug('completed the ``_list_images_needing_cached`` method')
        self.transientBucketIds = transientBucketIds
        return transientBucketIds, subtractedUrls, targetUrls, referenceUrls, tripletUrls

    def _download(
            self,
            transientBucketIds,
            subtractedUrls,
            targetUrls,
            referenceUrls,
            tripletUrls):
        """*cache the images for the survey under their transientBucketId folders in the web-server cache*

        **Key Arguments**

        - ``transientBucketIds`` -- the list of transientBucketId for the transients needing images downloaded.
        - ``subtractedUrls`` -- the list of subtracted image urls (same length as transientBucketIds list).
        - ``targetUrls`` -- the list of target image urls (same length as transientBucketIds list).
        - ``referenceUrls`` -- the list of reference image urls (same length as transientBucketIds list).
        - ``tripletUrls`` -- the list of triplet image urls (same length as transientBucketIds list).


        **Return**

        - ``subtractedStatus`` -- status of the subtracted image download (0 = fail, 1 = success, 2 = does not exist)
        - ``targetStatus`` -- status of the target image download (0 = fail, 1 = success, 2 = does not exist)
        - ``referenceStatus`` -- status of the reference image download (0 = fail, 1 = success, 2 = does not exist)
        - ``tripletStatus`` -- status of the triplet image download (0 = fail, 1 = success, 2 = does not exist)

        """
        self.log.debug('starting the ``_download`` method')

        downloadDirectoryPath = self.downloadDirectoryPath
        self.subtractedStatus = []
        self.targetStatus = []
        self.referenceStatus = []
        self.tripletStatus = []
        index = 1
        survey = self.survey.lower()

        # TOTAL TO DOWNLOAD
        count = len(transientBucketIds)

        # DOWNLOAD THE IMAGE SETS FOR EACH TRANSIENT AND ADD STATUS TO OUTPUT
        # ARRAYS. (0 = fail, 1 = success, 2 = does not exist)
        for tid, surl, turl, rurl, purl in zip(transientBucketIds, subtractedUrls, targetUrls, referenceUrls, tripletUrls):
            if index > 1:
                # Cursor up one line and clear line
                sys.stdout.write("\x1b[1A\x1b[2K")

            percent = (old_div(float(index), float(count))) * 100.
            print('%(index)s/%(count)s (%(percent)1.1f%% done): downloading %(survey)s stamps for transientBucketId: %(tid)s' % locals())

            statusArray = download_image_array(imageArray=[
                                               tid, surl, turl, rurl, purl], log=self.log, survey=self.survey, downloadPath=downloadDirectoryPath)
            self.subtractedStatus.append(statusArray[1])
            self.targetStatus.append(statusArray[2])
            self.referenceStatus.append(statusArray[3])
            self.tripletStatus.append(statusArray[4])
            index += 1

        self.log.debug('completed the ``_download`` method')
        return self.subtractedStatus, self.targetStatus, self.referenceStatus, self.tripletStatus

    def _update_database(
            self):
        """*update the database to show which images have been cached on the server*
        """
        self.log.debug('starting the ``_update_database`` method')

        if not len(self.tripletStatus):
            self.log.debug('completed the ``_update_database`` method')
            return None

        # ITERATE OVER 4 STAMP COLUMNS
        for imageType, column in list(self.stampFlagColumns.items()):
            if column:
                if imageType == "subtracted":
                    status = self.subtractedStatus
                if imageType == "target":
                    status = self.targetStatus
                if imageType == "reference":
                    status = self.referenceStatus
                if imageType == "triplet":
                    status = self.tripletStatus

                nonexist = []
                exist = []
                # NON-EXISTANT == STATUS 2
                nonexist[:] = [str(t) for t, s in zip(
                    self.transientBucketIds, status) if s == 2]
                nonexist = (",").join(nonexist)
                # EXISTANT == STATUS 1 (i.e. these are downloaded)
                exist[:] = [str(t) for t, s in zip(
                    self.transientBucketIds, status) if s == 1]
                exist = (",").join(exist)
                # GENERATE THE SQL TO UPDATE DATABASE
                sqlQuery = ""
                if len(nonexist):
                    sqlQuery += """update pesstoObjects set %(column)s = 3 where transientBucketId in (%(nonexist)s) and %(column)s = 2;""" % locals(
                    )
                    sqlQuery += """update pesstoObjects set %(column)s = 2 where transientBucketId in (%(nonexist)s) and (%(column)s is null or %(column)s = 0);""" % locals(
                    )

                    writequery(
                        log=self.log,
                        sqlQuery=sqlQuery,
                        dbConn=self.dbConn
                    )
                if len(exist):
                    sqlQuery = """update pesstoObjects set %(column)s = 1 where transientBucketId in (%(exist)s) and (%(column)s != 1 or %(column)s is null);""" % locals(
                    )

                    writequery(
                        log=self.log,
                        sqlQuery=sqlQuery,
                        dbConn=self.dbConn
                    )

        self.log.debug('completed the ``_update_database`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method


def download_image_array(
        imageArray,
        log,
        survey,
        downloadPath):
    """*download an array of transient image stamps*

    **Key Arguments**

    - ``log`` -- logger
    - ``imageArray`` -- [transientBucketId, subtractedUrl, targetUrl, referenceUrl, tripletUrl]
    - ``survey`` -- name of the survey to name stamps with
    - ``downloadPath`` -- directory to download the images into


    **Return**

    - statusArray -- [subtractedStatus, targetStatus, referenceStatus, tripletStatus]

    """
    tid = imageArray[0]
    statusArray = [imageArray[0]]

    # RECURSIVELY CREATE MISSING TRANSIENT DIRECTORIES
    downloadPath = "%(downloadPath)s/%(tid)s/" % locals()
    if not os.path.exists(downloadPath):
        os.makedirs(downloadPath)
    filepath = downloadPath + survey.lower()

    for url, stamp in zip(imageArray[1:], ["subtracted", "target", "reference", "triplet"]):
        if url:
            pathToWriteFile = "%(filepath)s_%(stamp)s_stamp.jpeg" % locals(
            )
        else:
            # NOTHING TO DOWNLOAD
            statusArray.append(0)
            continue

        try:
            response = requests.get(
                url=url,
                timeout=1.0
                # params={},
                # auth=HTTPBasicAuth('user', 'pwd')
            )
            content = response.content
            status_code = response.status_code
        except requests.exceptions.RequestException as e:
            if 'timed out' in str(e):
                print('timed out - try again next time' % locals())
                statusArray.append(0)
            else:
                print('HTTP Request failed - %(e)s' % locals())
                print("")
                statusArray.append(2)
            continue

        if status_code == 404:
            print('image not found' % locals())
            statusArray.append(2)
            continue

        # WRITE STAMP TO FILE
        try:
            writeFile = codecs.open(
                pathToWriteFile, mode='wb')
        except IOError as e:
            message = 'could not open the file %s' % (pathToWriteFile,)
            raise IOError(message)
        writeFile.write(content)
        writeFile.close()
        statusArray.append(1)

    return statusArray
